# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import uuid
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Optional

import pandas as pd
import pytest
from _pytest.monkeypatch import MonkeyPatch
from mlflow.entities import Run, RunData, RunInfo, RunStatus, RunTag
from mlflow.entities.lifecycle_stage import LifecycleStage
from mlflow.tracking.fluent import ActiveRun
from mlflow.utils.mlflow_tags import (
    MLFLOW_PARENT_RUN_ID,
    MLFLOW_RUN_NAME,
    MLFLOW_RUN_NOTE,
    MLFLOW_USER,
)


def _get_run_name_from_tags(tags):
    for tag in tags:
        if tag.key == MLFLOW_RUN_NAME:
            return tag.value


class MockMLFlowClient(object):
    def __init__(
        self, tracking_uri: Optional[str] = None, registry_uri: Optional[str] = None
    ):
        pass

    def download_artifacts(
        self, run_id: str, path: str, dst_path: Optional[str] = None
    ) -> str:
        if dst_path is None:
            dst_path = "tmp_unit_test"
        dst_path = str(Path(dst_path).absolute())
        dst_local_path = dst_path
        return dst_local_path

    def log_artifact(self, run_id: str, local_path: str, artifact_path=None) -> None:
        print("REACHED")
        return None

    def create_run(self, experiment_id, start_time=56020, tags=None, run_name=None):
        tags = tags if tags else {}
        user_id = tags.get(MLFLOW_USER, "unknown")
        experiment_id = 1 if experiment_id is None else experiment_id
        tags = tags or []
        run_name_tag = _get_run_name_from_tags(tags)
        if run_name and run_name_tag and run_name != run_name_tag:
            raise Exception(
                "Both 'run_name' argument and 'mlflow.runName' tag",
                "are specified, but with ",
                f"different values (run_name='{run_name}',",
                f"mlflow.runName='{run_name_tag}').",
            )
        run_name = run_name or run_name_tag or "experiment12345"
        if not run_name_tag:
            tags.append(RunTag(key=MLFLOW_RUN_NAME, value=run_name))
        run_uuid = uuid.uuid4().hex
        artifact_uri = Path("/path/to/artifacts/") / run_uuid / "artifacts"

        run_info = RunInfo(
            run_uuid=run_uuid,
            run_id=run_uuid,
            run_name=run_name,
            experiment_id=experiment_id,
            artifact_uri=artifact_uri,
            user_id=user_id,
            status=RunStatus.to_string(RunStatus.RUNNING),
            start_time=start_time,
            end_time=None,
            lifecycle_stage=LifecycleStage.ACTIVE,
        )
        return Run(run_info, RunData(None, None, tags))


def mock_start_run(
    run_id: str = None,
    experiment_id: Optional[str] = None,
    run_name: Optional[str] = None,
    nested: bool = False,
    tags: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
) -> ActiveRun:
    client = MockMLFlowClient()
    parent_run_id = None

    exp_id_for_run = experiment_id if experiment_id is not None else "experiment12345"
    user_specified_tags = deepcopy(tags) or {}
    if description:
        if MLFLOW_RUN_NOTE in user_specified_tags:
            raise Exception(
                f"Description is already set via the tag {MLFLOW_RUN_NOTE} in tags."
                f"Remove the key {MLFLOW_RUN_NOTE} from the tags or omit",
                "the description.",
            )
        user_specified_tags[MLFLOW_RUN_NOTE] = description
    if parent_run_id is not None:
        user_specified_tags[MLFLOW_PARENT_RUN_ID] = parent_run_id
    if run_name:
        user_specified_tags[MLFLOW_RUN_NAME] = run_name

    resolved_tags = user_specified_tags

    active_run_obj = client.create_run(
        experiment_id=exp_id_for_run, tags=resolved_tags, run_name=run_name
    )
    return ActiveRun(active_run_obj)


def mock_log_artifact(local_path: str, artifact_path: Optional[str] = None) -> None:
    run_id = "ex123456"
    MockMLFlowClient().log_artifact(run_id, local_path, artifact_path)


@pytest.fixture
def mlflow_client(monkeypatch: MonkeyPatch):
    import mlflow

    monkeypatch.setattr(mlflow.tracking, "MlflowClient", MockMLFlowClient)
    monkeypatch.setattr(mlflow.tracking.fluent, "start_run", mock_start_run)
    monkeypatch.setattr(mlflow, "log_artifact", mock_log_artifact)


@pytest.mark.parametrize(
    "run_id",
    ["ex123", "ex456", "ex621"],
)
@pytest.mark.parametrize(
    "artifact_path",
    ["path/to/file", "path/to/other_file"],
)
@pytest.mark.parametrize(
    "destination_path",
    ["path/to/file", "path/to/other/file"],
)
def test_download_all_artifacts_in_run(
    mlflow_client, run_id, artifact_path, destination_path
) -> None:
    from dioptra_builtins.artifacts.mlflow import download_all_artifacts_in_run

    dst_path = download_all_artifacts_in_run(run_id, artifact_path, destination_path)
    assert isinstance(dst_path, str)
    assert Path(destination_path) == Path(dst_path).relative_to(Path.cwd())


@pytest.mark.parametrize(
    "data_frame",
    [
        pd.DataFrame(
            [
                ["dollars", -10],
                ["books", 20],
                ["bottles", 0],
                ["frogs", -3],
                ["flies", 9],
                ["snakes", 60],
            ],
            columns=["things", "counts"],
        )
    ],
)
@pytest.mark.parametrize(
    ("file_name", "file_format", "output"),
    [
        ("pddf.csv", "csv.gz", "csv.gz"),
        ("pddf.csv", "csv", "csv"),
        ("pddf.csv", "csv.bz2", "csv.bz2"),
        ("pddf.csv", "csv.xz", "csv.xz"),
        ("pddf.json", "json", "json"),
        ("pddf.feather", "feather", "feather"),
        ("pddf.pkl", "pickle", "pkl"),
    ],
)
def test_upload_data_frame_artifact(
    mlflow_client, data_frame, file_name, file_format, output, tmp_path
) -> None:
    from dioptra_builtins.artifacts.mlflow import upload_data_frame_artifact

    upload_data_frame_artifact(data_frame, file_name, file_format, None, tmp_path)

    pwd = Path("." if tmp_path is None else tmp_path).absolute()
    assert (pwd / Path(file_name).with_suffix(f".{output}")).is_file()


@pytest.mark.parametrize(
    "source_dir",
    ["./tmp"],
)
@pytest.mark.parametrize(
    "tarball_filename",
    ["workflow.tar.gz"],
)
@pytest.mark.parametrize(
    "tarball_write_mode",
    ["w:gz"],
)
@pytest.mark.parametrize(
    "working_dir",
    [None],
)
def test_upload_directory_as_tarball_artifact(
    mlflow_client, source_dir, tarball_filename, tarball_write_mode, working_dir
) -> None:
    from dioptra_builtins.artifacts.mlflow import upload_directory_as_tarball_artifact

    Path(source_dir).mkdir(parents=True, exist_ok=True)
    fp = open(Path(source_dir) / "myfile.py", "w+")
    fp.close()
    Path(source_dir).mkdir(parents=True, exist_ok=True)
    fp = open(Path(source_dir) / "myfile2.py", "w+")
    fp.close()
    upload_directory_as_tarball_artifact(
        source_dir, tarball_filename, tarball_write_mode, working_dir
    )
    pwd = Path("." if working_dir is None else working_dir).absolute()
    assert (pwd / tarball_filename).is_file()


@pytest.mark.parametrize(
    ("working_dir", "file"),
    [
        ("./tmp_unit_test", "this.py"),
        ("./tmp_unit_tests", "test2.tar.gz"),
        ("./tmp_unit_test", "MLProject"),
    ],
)
def test_upload_file_as_artifact(mlflow_client, working_dir, file) -> None:
    from dioptra_builtins.artifacts.mlflow import upload_file_as_artifact

    Path(working_dir).mkdir(parents=True, exist_ok=True)
    fp = open(Path(working_dir) / file, "w+")
    fp.close()
    upload_file_as_artifact(Path(working_dir) / file)



from __future__ import with_statement

import logging
import os
from logging.config import fileConfig

from alembic import context
from sqlalchemy import create_engine

from dioptra.restapi.db.custom_types import GUID, TZDateTime

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
from flask import current_app

config.set_main_option(
    "sqlalchemy.url",
    str(current_app.extensions["migrate"].db.engine.url).replace("%", "%%"),
)
target_metadata = current_app.extensions["migrate"].db.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_url():
    app_env = os.getenv("DIOPTRA_RESTAPI_ENV", "dev")
    env_urls = {
        "dev": os.getenv("DIOPTRA_RESTAPI_DEV_DATABASE_URI"),
        "prod": os.getenv("DIOPTRA_RESTAPI_DATABASE_URI"),
    }

    return env_urls[app_env]


def render_item(type_, obj, autogen_context):
    """Apply custom rendering for selected items."""

    if type_ == "type" and isinstance(obj, GUID):
        # add import for this type
        autogen_context.imports.add("from dioptra.restapi.db.custom_types import GUID")
        return "%r" % obj

    if type_ == "type" and isinstance(obj, TZDateTime):
        # add import for this type
        autogen_context.imports.add(
            "from dioptra.restapi.db.custom_types import TZDateTime"
        )
        return "%r" % obj

    # default rendering for other objects
    return False


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = get_url()
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        render_item=render_item,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    # this callback is used to prevent an auto-migration from being generated
    # when there are no changes to the schema
    # reference: http://alembic.zzzcomputing.com/en/latest/cookbook.html
    def process_revision_directives(context, revision, directives):
        if getattr(config.cmd_opts, "autogenerate", False):
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []
                logger.info("No changes in schema detected.")

    connectable = create_engine(get_url())

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            process_revision_directives=process_revision_directives,
            render_item=render_item,
            **current_app.extensions["migrate"].configure_args,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import structlog
from structlog.stdlib import BoundLogger

from .image_data import ImageData

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    import tensorflow as tf
    from tensorflow import DType, Tensor
    from tensorflow.image import ResizeMethod

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


class TensorflowImageData(ImageData):
    def __init__(
        self,
        num_channels: int,
        expand_animations: bool = False,
        resize_method: str = ResizeMethod.BILINEAR,
        resize_preserve_aspect_ratio: bool = False,
        resize_antialias: bool = False,
        dtype: DType = tf.float32,
    ):
        self._num_channels = num_channels
        self._expand_animations = expand_animations
        self._resize_method = resize_method
        self._resize_preserve_aspect_ratio = resize_preserve_aspect_ratio
        self._resize_antialias = resize_antialias
        self._dtype = dtype

    def read_file(self, filepath: Tensor) -> Tensor:
        image = tf.io.read_file(filename=filepath)
        image = tf.io.decode_image(
            contents=image,
            channels=self._num_channels,
            expand_animations=self._expand_animations,
        )

        return tf.cast(image, self._dtype)

    def resize(self, images: Tensor, size: Tensor) -> Tensor:
        return tf.image.resize(
            images=images,
            size=size,
            method=self._resize_method,
            preserve_aspect_ratio=self._resize_preserve_aspect_ratio,
            antialias=self._resize_antialias,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

from typing import Any, Callable, Dict, Optional, Union

import numpy as np
import structlog
from structlog.stdlib import BoundLogger

from dioptra.sdk.exceptions import (
    EstimatorPredictGenericPredTypeError,
    TensorflowDependencyError,
)
from dioptra.sdk.generics import estimator_predict
from dioptra.sdk.utilities.decorators import require_package

LOGGER: BoundLogger = structlog.stdlib.get_logger()

try:
    from tensorflow.keras import Model

except ImportError:  # pragma: nocover
    LOGGER.warn(
        "Unable to import one or more optional packages, functionality may be reduced",
        package="tensorflow",
    )


@estimator_predict.register
def _(estimator: Model, x: Any, pred_type: str, **kwargs) -> np.ndarray:
    LOGGER.info(
        "Dispatch generic function",
        generic="estimator_predict",
        estimator="tensorflow.keras.Model",
        args_signature=("Model", "Any", "str"),
    )

    predict: Callable[..., np.ndarray] = dict(
        prob=keras_model_predict_proba,
        label=keras_model_predict_label,
    ).get(pred_type, _null_predict)

    try:
        prediction: np.ndarray = predict(
            estimator=estimator,
            x=x,
            **kwargs,
        )

    except EstimatorPredictGenericPredTypeError as exc:
        LOGGER.exception(
            "Unknown pred_type argument for estimator_predict function.",
            estimator="tensorflow.keras.Model",
            pred_type=pred_type,
        )
        raise exc

    return prediction


@require_package("tensorflow", exc_type=TensorflowDependencyError)
def keras_model_predict_proba(
    estimator: Model,
    x: Any,
    batch_size: Optional[int] = None,
    **kwargs,
) -> np.ndarray:
    predict_kwargs: Dict[str, Any] = dict(batch_size=batch_size, **kwargs)
    prediction: np.ndarray = estimator.predict(
        x=x,
        **{k: v for k, v in predict_kwargs.items() if v is not None},
    )

    return prediction


@require_package("tensorflow", exc_type=TensorflowDependencyError)
def keras_model_predict_label(
    estimator: Model,
    x: Any,
    batch_size: Optional[int] = None,
    **kwargs,
) -> np.ndarray:
    predict_kwargs: Dict[str, Any] = dict(batch_size=batch_size, **kwargs)
    prediction: np.ndarray = estimator.predict(
        x=x,
        **{k: v for k, v in predict_kwargs.items() if v is not None},
    )

    if prediction.shape[1] > 1:
        labels: Union[np.integer, np.ndarray] = np.argmax(prediction, axis=1)

        if isinstance(labels, np.integer):
            raise RuntimeError(
                "Label prediction should return an array, not an integer."
            )

        return labels

    flattened_labels: np.ndarray = (prediction.flatten() >= 0.5).astype(int)
    return flattened_labels


def _null_predict(*args, **kwargs) -> np.ndarray:
    raise EstimatorPredictGenericPredTypeError



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path
from string import Template
from typing import Final

import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.utils import read_text_file

LOGGER: BoundLogger = structlog.stdlib.get_logger()

SCRIPT_FILE_ENCODING: Final[str] = "utf-8"
CURRENT_PACKAGE: Final[str] = "dioptra.restapi.v1.workflows.lib"
TEMPLATE_FILENAME: Final[str] = "run_dioptra_job.py.tmpl"


def export_run_dioptra_job_script(
    job_id: int,
    experiment_id: int,
    task_engine_yaml_path: str,
    job_params_json_path: str,
    base_dir: Path,
    logger: BoundLogger | None = None,
) -> Path:
    """Render and export the run_dioptra_job.py script to a specified directory.

    Args:
        base_dir: The directory to export the rendered script to.
        logger: A structlog logger object to use for logging. A new logger will be
            created if None.

    Returns:
        The path to the exported script.
    """
    log = logger or LOGGER.new()  # noqa: F841
    script_path = Path(base_dir, "run_dioptra_job").with_suffix(".py")
    script = _render_run_dioptra_job_script(
        job_id=job_id,
        experiment_id=experiment_id,
        task_engine_yaml_path=task_engine_yaml_path,
        job_params_json_path=job_params_json_path,
    )

    with script_path.open("wt", encoding=SCRIPT_FILE_ENCODING) as f:
        f.write(script)

    return script_path


def _render_run_dioptra_job_script(
    job_id: int,
    experiment_id: int,
    task_engine_yaml_path: str,
    job_params_json_path: str,
):
    template_text = read_text_file(package=CURRENT_PACKAGE, filename=TEMPLATE_FILENAME)
    template = Template(template_text)

    return template.substitute(
        job_id=job_id,
        experiment_id=experiment_id,
        task_engine_yaml_path=task_engine_yaml_path,
        job_params_json_path=job_params_json_path,
    )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import importlib
import os
import tarfile
import tempfile
import uuid
from pathlib import Path
from typing import Final, cast
from urllib.parse import urlencode, urlparse, urlunparse

import requests
import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.routes import (
    V1_AUTH_ROUTE,
    V1_EXPERIMENTS_ROUTE,
    V1_JOBS_ROUTE,
    V1_ROOT,
    V1_WORKFLOWS_ROUTE,
)
from dioptra.sdk.utilities.contexts import sys_path_dirs
from dioptra.sdk.utilities.paths import set_cwd

LOGGER: BoundLogger = structlog.stdlib.get_logger()

ENV_DIOPTRA_API: Final[str] = "DIOPTRA_API"
ENV_DIOPTRA_WORKER_USERNAME: Final[str] = "DIOPTRA_WORKER_USERNAME"
ENV_DIOPTRA_WORKER_PASSWORD: Final[str] = "DIOPTRA_WORKER_PASSWORD"
ENV_MLFLOW_S3_ENDPOINT_URL: Final[str] = "MLFLOW_S3_ENDPOINT_URL"

DOWNLOAD_CHUNK_SIZE: Final[int] = 10 * 1024
TAR_GZ_FILE_TYPE: Final[str] = "tar_gz"
TAR_GZ_EXTENSION: Final[str] = ".tar.gz"

JOB_FILES_DOWNLOAD_ENDPOINT: Final[str] = f"{V1_WORKFLOWS_ROUTE}/jobFilesDownload"


class SimpleDioptraClient(object):
    def __init__(self, username: str, password: str, api_url: str):
        self._api_scheme, self._api_netloc = self._extract_scheme_and_netloc(api_url)
        self._username = username
        self._password = password
        self._session: requests.Session | None = None

    @property
    def session(self) -> requests.Session:
        if self._session is None:
            self.login()

        return cast(requests.Session, self._session)

    def login(self) -> None:
        if self._session is None:
            self._session = requests.Session()

        response = self._session.post(
            self._build_url(f"{V1_AUTH_ROUTE}/login"),
            json={"username": self._username, "password": self._password},
        )
        response.raise_for_status()

    def download_job_files(self, job_id: int, output_dir: Path) -> Path:
        url = self._build_url(
            JOB_FILES_DOWNLOAD_ENDPOINT,
            query_params={"jobId": str(job_id), "fileType": TAR_GZ_FILE_TYPE},
        )
        job_files_path = (output_dir / "job_files").with_suffix(TAR_GZ_EXTENSION)
        with (
            self.session.get(url, stream=True) as response,
            job_files_path.open(mode="wb") as f,
        ):
            for chunk in response.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                f.write(chunk)

        return job_files_path

    def set_job_status(self, job_id: int, experiment_id: int, status: str) -> None:
        url = self._build_url(
            f"{V1_EXPERIMENTS_ROUTE}/{experiment_id}/jobs/{job_id}/status"
        )
        response = self.session.put(url, json={"status": status})
        response.raise_for_status()

    def set_job_mlflow_run_id(
        self, job_id: int, mlflow_run_id: str | uuid.UUID
    ) -> None:
        url = self._build_url(f"{V1_JOBS_ROUTE}/{job_id}/mlflowRun")
        payload = {
            "mlflowRunId": (
                mlflow_run_id.hex
                if isinstance(mlflow_run_id, uuid.UUID)
                else mlflow_run_id
            )
        }
        response = self.session.post(url, json=payload)
        response.raise_for_status()

    def _build_url(
        self, endpoint: str, query_params: dict[str, str] | None = None
    ) -> str:
        query_params = query_params or {}

        return urlunparse(
            (
                self._api_scheme,
                self._api_netloc,
                f"/{V1_ROOT}/{endpoint}",
                "",
                urlencode(query_params),
                "",
            )
        )

    @staticmethod
    def _extract_scheme_and_netloc(api_url: str) -> tuple[str, str]:
        parsed_api_url = urlparse(url=api_url)
        return parsed_api_url.scheme, parsed_api_url.netloc


def run_v1_dioptra_job(job_id: int, experiment_id: int) -> None:  # noqa: C901
    """Fetches the job files from the Dioptra API and runs the job.

    Args:
        job_id: The Dioptra job ID.
        experiment_id: The Dioptra experiment ID.
    """
    log = LOGGER.new(job_id=job_id, experiment_id=experiment_id)  # noqa: F841

    if (username := os.getenv(ENV_DIOPTRA_WORKER_USERNAME)) is None:
        log.error(f"{ENV_DIOPTRA_WORKER_USERNAME} environment variable is not set")
        raise ValueError(
            f"{ENV_DIOPTRA_WORKER_USERNAME} environment variable is not set"
        )

    if (password := os.getenv(ENV_DIOPTRA_WORKER_PASSWORD)) is None:
        log.error(f"{ENV_DIOPTRA_WORKER_PASSWORD} environment variable is not set")
        raise ValueError(
            f"{ENV_DIOPTRA_WORKER_PASSWORD} environment variable is not set"
        )

    if (api_url := os.getenv(ENV_DIOPTRA_API)) is None:
        log.error(f"{ENV_DIOPTRA_API} environment variable is not set")
        raise ValueError(f"{ENV_DIOPTRA_API} environment variable is not set")

    # Instantiate a Dioptra client and login using worker's authentication details
    client = SimpleDioptraClient(username=username, password=password, api_url=api_url)

    # Set Dioptra Job status to "started"
    client.set_job_status(job_id=job_id, experiment_id=experiment_id, status="started")

    if os.getenv(ENV_MLFLOW_S3_ENDPOINT_URL) is None:
        client.set_job_status(
            job_id=job_id, experiment_id=experiment_id, status="failed"
        )
        log.error(f"{ENV_MLFLOW_S3_ENDPOINT_URL} environment variable is not set")
        raise ValueError(
            f"{ENV_MLFLOW_S3_ENDPOINT_URL} environment variable is not set"
        )

    # Set up a temporary directory and set it as the current working directory
    with tempfile.TemporaryDirectory() as tempdir, set_cwd(tempdir):
        working_dir = Path(tempdir)
        plugins_dir = working_dir / "plugins"
        run_dioptra_job_path = working_dir / "run_dioptra_job.py"

        # Use client to download the job files for the provided job_id
        try:
            job_files_package = client.download_job_files(
                job_id=job_id, output_dir=working_dir
            )

        except Exception as e:
            client.set_job_status(
                job_id=job_id, experiment_id=experiment_id, status="failed"
            )
            log.exception("Could not download job files")
            raise e

        # Unpack the (trusted) tar.gz file in it.
        try:
            with tarfile.open(job_files_package, mode="r:*") as tar:
                tar.extractall(path=working_dir, filter="data")

        except Exception as e:
            client.set_job_status(
                job_id=job_id, experiment_id=experiment_id, status="failed"
            )
            log.exception("Could not extract from tar file")
            raise e

        # Import the run_dioptra_job.py file as a module
        try:
            with sys_path_dirs(dirs=(str(working_dir),)):
                run_dioptra_job = importlib.import_module(run_dioptra_job_path.stem)

        except Exception as e:
            client.set_job_status(
                job_id=job_id, experiment_id=experiment_id, status="failed"
            )
            log.exception("Could not import run_dioptra_job.py")
            raise e

        # Execute the main function in the included script file.
        try:
            run_dioptra_job.main(
                plugins_dir=plugins_dir,
                enable_mlflow_tracking=True,
                dioptra_client=client,
                logger=log,
            )
            client.set_job_status(
                job_id=job_id, experiment_id=experiment_id, status="finished"
            )

        except Exception as e:
            client.set_job_status(
                job_id=job_id, experiment_id=experiment_id, status="failed"
            )
            log.exception("Error running job")
            raise e



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""A module for registering the error handlers for the application.

.. |Api| replace:: :py:class:`flask_restx.Api`
"""
from __future__ import annotations

import http
import typing

import structlog
from flask import request
from flask_restx import Api
from structlog.stdlib import BoundLogger

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def add_attribute_values(**kwargs: typing.Any) -> list[str]:
    """
    Helper function to add attribute name/value pairs to a list for use in an error
    message.

    Args:
        buffer: The StringIO instance into which the attribute name/value pairs are to
        be added.
        attributes: the list of attribute value pairs to add to the buffer.
    """
    length = len(kwargs)

    def sep(index: int) -> str:
        if index == 0:
            return " with"

        if index == length - 1:
            return ", and"

        return ","

    return [
        f"{sep(index)} {key} having value ({value})"
        for index, (key, value) in enumerate(kwargs.items())
    ]


class DioptraError(Exception):
    """
    Generic Dioptra Error.
    Args:
        message: An error specific message to display that provide context for why the
        error was raised.
    """

    def __init__(self, message: str):
        self.message: str = message

    def to_message(self) -> str:
        if self.__cause__ is None:
            return self.message

        if isinstance(self.__cause__, DioptraError):
            return f"{self.message} Cause: {self.__cause__.to_message()}"

        return f"{self.message} Cause: {self.__cause__}"


class EntityDoesNotExistError(DioptraError):
    """
    The requested entity does not exist.
    Args:
        entity_type: the entity type name (e.g. "group" or "queue")
        kwargs: the attribute value pairs used to request the entity
    """

    def __init__(self, entity_type: str | None = None, **kwargs: typing.Any):
        super().__init__(
            "".join(
                [
                    "Failed to locate ",
                    "an entity" if entity_type is None else entity_type,
                    *add_attribute_values(**kwargs),
                    ".",
                ]
            )
        )
        self.entity_type = "unknown" if entity_type is None else entity_type
        self.entity_attributes = kwargs


class EntityExistsError(DioptraError):
    """
    The requested entity already exists.
    Args:
        entity_type: the entity type name (e.g. "group" or "queue")
        existing_id: the id of the existing entity
        kwargs: the attribute value pairs used to request the entity
    """

    def __init__(self, entity_type: str, existing_id: int, **kwargs: typing.Any):
        super().__init__(
            "".join(
                [
                    f"The {entity_type}",
                    *add_attribute_values(**kwargs),
                    " is not available.",
                ]
            )
        )
        self.entity_type = entity_type
        self.entity_attributes = kwargs
        self.existing_id = existing_id


class LockError(DioptraError):
    """
    Top-level Lock Error.

    Args:
        message: a message describing the lock error
    """

    def __init__(self, message: str):
        super().__init__(message)


class ReadOnlyLockError(LockError):
    """The type has a read-only lock and cannot be modified."""

    def __init__(self, type: str, **kwargs: typing.Any):
        super().__init__(
            "".join(
                [
                    f"The {type} type",
                    *add_attribute_values(**kwargs),
                    " has a read-only lock and cannot be modified.",
                ]
            )
        )
        self.entity_type = type
        self.entity_attributes = kwargs


class BackendDatabaseError(DioptraError):
    """The backend database returned an unexpected response."""

    def __init__(self):
        super().__init__(
            "The backend database returned an unexpected response, please "
            "contact the system administrator."
        )


class SearchNotImplementedError(DioptraError):
    """The search functionality has not been implemented."""

    def __init__(self):
        super().__init__("The search functionality has not been implemented.")


class SearchParseError(DioptraError):
    """The search query could not be parsed."""

    def __init__(self, context: str, error: str):
        super().__init__("The provided search query could not be parsed.")
        self.context = context
        self.error = error


class DraftDoesNotExistError(DioptraError):
    """The requested draft does not exist."""

    def __init__(self, **kwargs: typing.Any):
        super().__init__(
            "".join(
                [
                    "The requested draft",
                    *add_attribute_values(**kwargs),
                    " does not exist.",
                ]
            )
        )
        self.entity_attributes = kwargs


class DraftAlreadyExistsError(DioptraError):
    """The draft already exists."""

    def __init__(self, type: str, id: int):
        super().__init__(f"A draft for a [{type}] with id: {id} already exists.")
        self.resource_type = type
        self.resource_id = id


class SortParameterValidationError(DioptraError):
    """The sort parameters are not valid."""

    def __init__(self, type: str, column: str, **kwargs):
        super().__init__(f"The sort parameter, {column}, for {type} is not sortable.")


class QueryParameterValidationError(DioptraError):
    """Input parameters failed validation."""

    def __init__(self, type: str, constraint: str, **kwargs):
        super().__init__(
            "".join(
                [
                    f"Input parameters for {type} failed {constraint} check",
                    *add_attribute_values(**kwargs),
                    ".",
                ]
            )
        )
        self.resource_type = type
        self.constraint = constraint
        self.parameters = kwargs


class QueryParameterNotUniqueError(QueryParameterValidationError):
    """Query Parameters failed unique validatation check."""

    def __init__(self, type: str, **kwargs):
        super().__init__(type, "unique", **kwargs)


class JobInvalidStatusTransitionError(DioptraError):
    """The requested status transition is invalid."""

    def __init__(self):
        super().__init__("The requested job status update is invalid.")


class JobInvalidParameterNameError(DioptraError):
    """The requested job parameter name is invalid."""

    def __init__(self):
        super().__init__(
            "A provided job parameter name does not match any entrypoint " "parameters."
        )


class JobMlflowRunAlreadySetError(DioptraError):
    """The requested job already has an mlflow run id set."""

    def __init__(self):
        super().__init__(
            "The requested job already has an mlflow run id set. It may "
            "not be changed."
        )


class EntityDependencyError(DioptraError):
    """
    Base Error for dependency problems between entities.

    Args:
        message: a message describing the dependecy error
    """

    def __init__(self, message: str):
        super().__init__(message)


class EntityNotRegisteredError(DioptraError):
    """
    An entity could not be located based on a relationship with another entity.

    Args:
        parent_type: the parent or owning type of the relation
        parent_id: the parent or owning id of the relation
        child_type: the child or dependent type of the relation
        child_id: the child or dependent id of the relation
    """

    def __init__(
        self, parent_type: str, parent_id: int, child_type: str, child_id: int
    ):
        super().__init__(
            f"The requested {child_type} having id ({child_id}) is not registered to "
            f"the {parent_type} having id ({parent_id})."
        )
        self.parent_type = parent_type
        self.parent_id = parent_id
        self.child_type = child_type
        self.child_id = child_id


class PluginParameterTypeMatchesBuiltinTypeError(DioptraError):
    """The plugin parameter type name cannot match a built-in type."""

    def __init__(self):
        super().__init__(
            "The requested plugin parameter type name matches a built-in "
            "type. Please select another and resubmit."
        )


# User Errors
class NoCurrentUserError(DioptraError):
    """There is no currently logged-in user."""

    def __init__(self):
        super().__init__("There is no currently logged-in user.")


class UserPasswordChangeError(DioptraError):
    """Password change failed."""

    def __init__(self, message: str):
        super().__init__(message)


class UserPasswordError(DioptraError):
    """Password Error."""

    def __init__(self, message: str):
        super().__init__(message)


def error_result(
    error: DioptraError, status: http.HTTPStatus, detail: dict[str, typing.Any]
) -> tuple[dict[str, typing.Any], int]:
    return {
        "error": error.__class__.__name__,
        "message": f"{status.phrase} - {error.message}",
        "detail": detail,
        "originating_path": request.full_path,
    }, status.value


# Silenced Complexity error for this function since it is a straitfoward registration of
# error handlers
def register_error_handlers(api: Api, **kwargs) -> None:  # noqa: C901
    """Registers the error handlers with the main application.

    Args:
        api: The main REST |Api| object.
    """
    log: BoundLogger = kwargs.get("log", LOGGER.new())

    @api.errorhandler(EntityDoesNotExistError)
    def handle_resource_does_not_exist_error(error: EntityDoesNotExistError):
        log.debug(
            "Entity not found", entity_type=error.entity_type, **error.entity_attributes
        )
        return error_result(
            error,
            http.HTTPStatus.NOT_FOUND,
            {"entity_type": error.entity_type, **error.entity_attributes},
        )

    @api.errorhandler(EntityExistsError)
    def handle_entity_exists_error(error: EntityExistsError):
        log.debug(
            "Entity exists",
            entity_type=error.entity_type,
            existing_id=error.existing_id,
            **error.entity_attributes,
        )
        return error_result(
            error,
            http.HTTPStatus.CONFLICT,
            {
                "entity_type": error.entity_type,
                "existing_id": error.existing_id,
                "entity_attributes": {**error.entity_attributes},
            },
        )

    @api.errorhandler(BackendDatabaseError)
    def handle_backend_database_error(error: BackendDatabaseError):
        log.error(error.to_message())
        return error_result(error, http.HTTPStatus.INTERNAL_SERVER_ERROR, {})

    @api.errorhandler(SearchNotImplementedError)
    def handle_search_not_implemented_error(error: SearchNotImplementedError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.NOT_IMPLEMENTED, {})

    @api.errorhandler(SearchParseError)
    def handle_search_parse_error(error: SearchParseError):
        log.debug(error.to_message())
        return error_result(
            error,
            http.HTTPStatus.UNPROCESSABLE_ENTITY,
            {"query": error.args[0], "reason": error.args[1]},
        )

    @api.errorhandler(DraftDoesNotExistError)
    def handle_draft_does_not_exist(error: DraftDoesNotExistError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.NOT_FOUND, {})

    @api.errorhandler(DraftAlreadyExistsError)
    def handle_draft_already_exists(error: DraftAlreadyExistsError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.BAD_REQUEST, {})

    @api.errorhandler(LockError)
    def handle_lock_error(error: LockError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.FORBIDDEN, {})

    @api.errorhandler(NoCurrentUserError)
    def handle_no_current_user_error(error: NoCurrentUserError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.UNAUTHORIZED, {})

    @api.errorhandler(UserPasswordChangeError)
    def handle_password_change_error(error: UserPasswordChangeError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.FORBIDDEN, {})

    @api.errorhandler(UserPasswordError)
    def handle_user_password_error(error: UserPasswordError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.UNAUTHORIZED, {})

    @api.errorhandler(DioptraError)
    def handle_base_error(error: DioptraError):
        log.debug(error.to_message())
        return error_result(error, http.HTTPStatus.BAD_REQUEST, {})



"""Remove legacy v0 tables

Revision ID: 4b2d781f8bb4
Revises: fd786b5377d6
Create Date: 2024-07-25 15:49:07.329979

"""

import sqlalchemy as sa
from alembic import op

from dioptra.restapi.db.custom_types import GUID

# revision identifiers, used by Alembic.
revision = "4b2d781f8bb4"
down_revision = "fd786b5377d6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.drop_index("ix_legacy_queues_name")

    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.drop_index("ix_legacy_jobs_experiment_id")
        batch_op.drop_index("ix_legacy_jobs_mlflow_run_id")
        batch_op.drop_index("ix_legacy_jobs_queue_id")
        batch_op.drop_index("ix_legacy_jobs_status")

    op.drop_table("legacy_jobs")
    op.drop_table("legacy_users")
    op.drop_table("legacy_queue_locks")
    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.drop_index("ix_legacy_experiments_name")

    op.drop_table("legacy_experiments")
    op.drop_table("legacy_queues")
    op.drop_table("legacy_job_statuses")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    legacy_job_statuses = op.create_table(
        "legacy_job_statuses",
        sa.Column("status", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("status", name="pk_legacy_job_statuses"),
    )

    op.bulk_insert(
        legacy_job_statuses,
        [
            {"status": "queued"},
            {"status": "started"},
            {"status": "deferred"},
            {"status": "finished"},
            {"status": "failed"},
        ],
    )

    op.create_table(
        "legacy_queues",
        sa.Column(
            "queue_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("queue_id", name="pk_legacy_queues"),
    )
    with op.batch_alter_table("legacy_queues", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_queues_name", ["name"], unique=1)

    op.create_table(
        "legacy_experiments",
        sa.Column(
            "experiment_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column("name", sa.Text(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("experiment_id", name="pk_legacy_experiments"),
    )
    with op.batch_alter_table("legacy_experiments", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_experiments_name", ["name"], unique=1)

    op.create_table(
        "legacy_queue_locks",
        sa.Column(
            "queue_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["queue_id"],
            ["legacy_queues.queue_id"],
            name="fk_legacy_queue_locks_queue_id_legacy_queues",
        ),
        sa.PrimaryKeyConstraint("queue_id", name="pk_legacy_queue_locks"),
    )
    op.create_table(
        "legacy_users",
        sa.Column(
            "user_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("username", sa.Text(), nullable=False),
        sa.Column("password", sa.Text(), nullable=False),
        sa.Column("email_address", sa.Text(), nullable=False),
        sa.Column("created_on", sa.DateTime(), nullable=False),
        sa.Column("last_modified_on", sa.DateTime(), nullable=False),
        sa.Column("last_login_on", sa.DateTime(), nullable=False),
        sa.Column("password_expire_on", sa.DateTime(), nullable=False),
        sa.Column("is_deleted", sa.Boolean(), nullable=False),
        sa.Column("alternative_id", GUID(), nullable=False),
        sa.PrimaryKeyConstraint("user_id", name="pk_legacy_users"),
        sa.UniqueConstraint("alternative_id", name="uq_legacy_users_alternative_id"),
    )
    op.create_table(
        "legacy_jobs",
        sa.Column("job_id", sa.String(length=36), nullable=False),
        sa.Column("mlflow_run_id", sa.String(length=36), nullable=True),
        sa.Column("experiment_id", sa.BigInteger(), nullable=True),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column("timeout", sa.Text(), nullable=True),
        sa.Column("workflow_uri", sa.Text(), nullable=True),
        sa.Column("entry_point", sa.Text(), nullable=True),
        sa.Column("entry_point_kwargs", sa.Text(), nullable=True),
        sa.Column("depends_on", sa.String(length=36), nullable=True),
        sa.Column("status", sa.String(length=255), nullable=True),
        sa.Column("queue_id", sa.BigInteger(), nullable=True),
        sa.ForeignKeyConstraint(
            ["experiment_id"],
            ["legacy_experiments.experiment_id"],
            name="fk_legacy_jobs_experiment_id_legacy_experiments",
        ),
        sa.ForeignKeyConstraint(
            ["queue_id"],
            ["legacy_queues.queue_id"],
            name="fk_legacy_jobs_queue_id_legacy_queues",
        ),
        sa.ForeignKeyConstraint(
            ["status"],
            ["legacy_job_statuses.status"],
            name="fk_legacy_jobs_status_legacy_job_statuses",
        ),
        sa.PrimaryKeyConstraint("job_id", name="pk_legacy_jobs"),
    )
    with op.batch_alter_table("legacy_jobs", schema=None) as batch_op:
        batch_op.create_index("ix_legacy_jobs_status", ["status"], unique=False)
        batch_op.create_index("ix_legacy_jobs_queue_id", ["queue_id"], unique=False)
        batch_op.create_index(
            "ix_legacy_jobs_mlflow_run_id", ["mlflow_run_id"], unique=False
        )
        batch_op.create_index(
            "ix_legacy_jobs_experiment_id", ["experiment_id"], unique=False
        )
    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Plugin resources."""
import uuid
from typing import cast

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import inject
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.routes import V1_PLUGIN_FILES_ROUTE, V1_PLUGINS_ROUTE
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import IdStatusResponseSchema
from dioptra.restapi.v1.shared.drafts.controller import (
    generate_nested_resource_drafts_endpoint,
    generate_nested_resource_drafts_id_endpoint,
    generate_nested_resource_id_draft_endpoint,
    generate_resource_drafts_endpoint,
    generate_resource_drafts_id_endpoint,
    generate_resource_id_draft_endpoint,
)
from dioptra.restapi.v1.shared.snapshots.controller import (
    generate_nested_resource_snapshots_endpoint,
    generate_nested_resource_snapshots_id_endpoint,
    generate_resource_snapshots_endpoint,
    generate_resource_snapshots_id_endpoint,
)
from dioptra.restapi.v1.shared.tags.controller import (
    generate_nested_resource_tags_endpoint,
    generate_nested_resource_tags_id_endpoint,
    generate_resource_tags_endpoint,
    generate_resource_tags_id_endpoint,
)

from .schema import (
    PluginFileGetQueryParameters,
    PluginFileMutableFieldsSchema,
    PluginFilePageSchema,
    PluginFileSchema,
    PluginGetQueryParameters,
    PluginMutableFieldsSchema,
    PluginPageSchema,
    PluginSchema,
)
from .service import (
    PLUGIN_FILE_RESOURCE_TYPE,
    PLUGIN_FILE_SEARCHABLE_FIELDS,
    PLUGIN_RESOURCE_TYPE,
    PLUGIN_SEARCHABLE_FIELDS,
    PluginIdFileIdService,
    PluginIdFileService,
    PluginIdService,
    PluginService,
)

LOGGER: BoundLogger = structlog.stdlib.get_logger()

api: Namespace = Namespace("Plugins", description="Plugins endpoint")


# WARNING: Do not move the PluginFile Snapshots sub-endpoint definitions.
# They must be declared first because of an issue related interactions with the schemas.
PluginFileSnapshotsResource = generate_nested_resource_snapshots_endpoint(
    api=api,
    resource_model=models.PluginFile,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    base_resource_route=V1_PLUGINS_ROUTE,
    searchable_fields=PLUGIN_FILE_SEARCHABLE_FIELDS,
    page_schema=PluginFilePageSchema,
    build_fn=utils.build_plugin_file,
)
PluginFileSnapshotsIdResource = generate_nested_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.PluginFile,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    response_schema=PluginFileSchema,
    build_fn=utils.build_plugin_file,
)


@api.route("/")
class PluginEndpoint(Resource):
    @inject
    def __init__(self, plugin_service: PluginService, *args, **kwargs) -> None:
        """Initialize the plugin resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_service: A PluginService object.
        """
        self._plugin_service = plugin_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=PluginGetQueryParameters, api=api)
    @responds(schema=PluginPageSchema, api=api)
    def get(self):
        """Gets a list of all Plugin resources."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Plugin", request_type="GET"
        )
        parsed_query_params = request.parsed_query_params  # noqa: F841

        group_id = parsed_query_params["group_id"]
        search_string = parsed_query_params["search"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        plugins, total_num_plugins = self._plugin_service.get(
            group_id=group_id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )
        return utils.build_paging_envelope(
            V1_PLUGINS_ROUTE,
            build_fn=utils.build_plugin,
            data=plugins,
            group_id=group_id,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_plugins,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @accepts(schema=PluginSchema, api=api)
    @responds(schema=PluginSchema, api=api)
    def post(self):
        """Creates a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Plugin", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # noqa: F841

        plugin = self._plugin_service.create(
            name=parsed_obj["name"],
            description=parsed_obj["description"],
            group_id=parsed_obj["group_id"],
            log=log,
        )
        return utils.build_plugin(plugin)


@api.route("/<int:id>")
@api.param("id", "ID for the Plugin resource.")
class PluginIdEndpoint(Resource):
    @inject
    def __init__(self, plugin_id_service: PluginIdService, *args, **kwargs) -> None:
        """Initialize the plugin resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_id_service: A PluginIdService object.
        """
        self._plugin_id_service = plugin_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=PluginSchema, api=api)
    def get(self, id: int):
        """Gets a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Plugin", request_type="GET", id=id
        )
        plugin = cast(
            models.Plugin,
            self._plugin_id_service.get(id, error_if_not_found=True, log=log),
        )
        return utils.build_plugin(plugin)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="Plugin",
            request_type="DELETE",
            id=id,
        )
        return self._plugin_id_service.delete(plugin_id=id, log=log)

    @login_required
    @accepts(schema=PluginMutableFieldsSchema, api=api)
    @responds(schema=PluginSchema, api=api)
    def put(self, id: int):
        """Modifies a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="Plugin", request_type="PUT", id=id
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        plugin = cast(
            models.Plugin,
            self._plugin_id_service.modify(
                id,
                name=parsed_obj["name"],
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_plugin(plugin)


@api.route("/<int:id>/files")
@api.param("id", "ID for the Plugin resource.")
class PluginIdFilesEndpoint(Resource):
    @inject
    def __init__(
        self, plugin_id_file_service: PluginIdFileService, *args, **kwargs
    ) -> None:
        """Initialize the plugin file resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_service: A PluginService object.
        """
        self._plugin_id_file_service = plugin_id_file_service
        super().__init__(*args, **kwargs)

    @login_required
    @accepts(query_params_schema=PluginFileGetQueryParameters, api=api)
    @responds(schema=PluginFilePageSchema, api=api)
    def get(self, id: int):
        """Gets a list of all PluginFile resources for a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginFile",
            request_type="GET",
            id=id,
        )
        parsed_query_params = request.parsed_query_params  # type: ignore # noqa: F841

        search_string = parsed_query_params["search"]
        page_index = parsed_query_params["index"]
        page_length = parsed_query_params["page_length"]
        sort_by_string = parsed_query_params["sort_by"]
        descending = parsed_query_params["descending"]

        plugin_files, total_num_plugin_files = self._plugin_id_file_service.get(
            plugin_id=id,
            search_string=search_string,
            page_index=page_index,
            page_length=page_length,
            sort_by_string=sort_by_string,
            descending=descending,
            log=log,
        )

        return utils.build_paging_envelope(
            f"plugins/{id}/files",
            utils.build_plugin_file,
            plugin_files,
            group_id=None,
            query=search_string,
            draft_type=None,
            index=page_index,
            length=page_length,
            total_num_elements=total_num_plugin_files,
            sort_by=sort_by_string,
            descending=descending,
        )

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int):
        """Deletes all PluginFile resources associated with a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginFile",
            request_type="DELETE",
            id=id,
        )
        return self._plugin_id_file_service.delete(plugin_id=id, log=log)

    @login_required
    @accepts(schema=PluginFileSchema(exclude=["groupId"]), api=api)
    @responds(schema=PluginFileSchema, api=api)
    def post(self, id: int):
        """Creates a PluginFile resource associated with a Plugin resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()), resource="PluginFile", request_type="POST"
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841

        plugin_file = self._plugin_id_file_service.create(
            filename=parsed_obj["filename"],
            contents=parsed_obj["contents"],
            description=parsed_obj["description"],
            tasks=parsed_obj["tasks"],
            plugin_id=id,
            log=log,
        )
        return utils.build_plugin_file(plugin_file)


@api.route("/<int:id>/files/<int:fileId>")
@api.param("id", "ID for the Plugin resource.")
@api.param("fileId", "ID for the PluginFile resource.")
class PluginIdFileIdEndpoint(Resource):
    @inject
    def __init__(
        self, plugin_file_id_service: PluginIdFileIdService, *args, **kwargs
    ) -> None:
        """Initialize the plugin file id resource.

        All arguments are provided via dependency injection.

        Args:
            plugin_file_id_service: PluginIdFileIdService
        """
        self._plugin_file_id_service = plugin_file_id_service
        super().__init__(*args, **kwargs)

    @login_required
    @responds(schema=PluginFileSchema, api=api)
    def get(self, id: int, fileId: int):
        """Gets a PluginFile resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginFile",
            request_type="GET",
            id=id,
        )
        plugin_file = cast(
            models.PluginFile,
            self._plugin_file_id_service.get(
                id, plugin_file_id=fileId, error_if_not_found=True, log=log
            ),
        )
        return utils.build_plugin_file(plugin_file)

    @login_required
    @responds(schema=IdStatusResponseSchema, api=api)
    def delete(self, id: int, fileId: int):
        """Deletes a PluginFile resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginFile",
            request_type="DELETE",
            id=id,
        )
        return self._plugin_file_id_service.delete(id, plugin_file_id=fileId, log=log)

    @login_required
    @accepts(schema=PluginFileMutableFieldsSchema, api=api)
    @responds(schema=PluginFileSchema, api=api)
    def put(self, id: int, fileId: int):
        """Modifies a PluginFile resource."""
        log = LOGGER.new(
            request_id=str(uuid.uuid4()),
            resource="PluginFile",
            request_type="PUT",
            id=id,
        )
        parsed_obj = request.parsed_obj  # type: ignore # noqa: F841
        plugin_file = cast(
            models.PluginFile,
            self._plugin_file_id_service.modify(
                id,
                plugin_file_id=fileId,
                filename=parsed_obj["filename"],
                contents=parsed_obj["contents"],
                tasks=parsed_obj["tasks"],
                description=parsed_obj["description"],
                error_if_not_found=True,
                log=log,
            ),
        )
        return utils.build_plugin_file(plugin_file)


PluginDraftResource = generate_resource_drafts_endpoint(
    api,
    resource_name=PLUGIN_RESOURCE_TYPE,
    route_prefix=V1_PLUGINS_ROUTE,
    request_schema=PluginSchema,
)
PluginDraftIdResource = generate_resource_drafts_id_endpoint(
    api,
    resource_name=PLUGIN_RESOURCE_TYPE,
    request_schema=PluginMutableFieldsSchema,
)
PluginIdDraftResource = generate_resource_id_draft_endpoint(
    api,
    resource_name=PLUGIN_RESOURCE_TYPE,
    request_schema=PluginMutableFieldsSchema,
)

PluginFileDraftResource = generate_nested_resource_drafts_endpoint(
    api,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    base_resource_route=V1_PLUGINS_ROUTE,
    request_schema=PluginFileSchema(exclude=["groupId"]),
)
PluginFileDraftIdResource = generate_nested_resource_drafts_id_endpoint(
    api,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    request_schema=PluginFileSchema(exclude=["groupId"]),
)
PluginFileIdDraftResource = generate_nested_resource_id_draft_endpoint(
    api,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    request_schema=PluginFileSchema(exclude=["groupId"]),
)

PluginSnapshotsResource = generate_resource_snapshots_endpoint(
    api=api,
    resource_model=models.Plugin,
    resource_name=PLUGIN_RESOURCE_TYPE,
    route_prefix=V1_PLUGINS_ROUTE,
    searchable_fields=PLUGIN_SEARCHABLE_FIELDS,
    page_schema=PluginPageSchema,
    build_fn=utils.build_plugin,
)
PluginSnapshotsIdResource = generate_resource_snapshots_id_endpoint(
    api=api,
    resource_model=models.Plugin,
    resource_name=PLUGIN_RESOURCE_TYPE,
    response_schema=PluginSchema,
    build_fn=utils.build_plugin,
)

PluginTagsResource = generate_resource_tags_endpoint(
    api=api,
    resource_name=PLUGIN_RESOURCE_TYPE,
)
PluginTagsIdResource = generate_resource_tags_id_endpoint(
    api=api,
    resource_name=PLUGIN_RESOURCE_TYPE,
)
PluginFileTagsIdResource = generate_nested_resource_tags_endpoint(
    api=api,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    base_resource_name=PLUGIN_RESOURCE_TYPE,
)
PluginFileTagsIdResource = generate_nested_resource_tags_id_endpoint(
    api=api,
    resource_name=PLUGIN_FILE_RESOURCE_TYPE,
    resource_route=V1_PLUGIN_FILES_ROUTE,
    base_resource_name=PLUGIN_RESOURCE_TYPE,
)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import io
import os
from pathlib import Path
from typing import BinaryIO, List, Optional, Union

import pytest
import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.v1.shared.io_file_service import (
    IOFileService,
    UnsafeArchiveMemberPath,
)
from tests.utils import make_tarball_bytes

LOGGER: BoundLogger = structlog.stdlib.get_logger()


@pytest.fixture
def io_file_service(dependency_injector) -> IOFileService:
    return dependency_injector.get(IOFileService)


def test_safe_extract_archive(
    io_file_service: IOFileService,
    task_plugin_archive: BinaryIO,
    tmp_path: Path,
) -> None:
    output_dir: Path = tmp_path / "extracted_plugins"
    output_dir.mkdir()

    extracted_files: List[str] = io_file_service.safe_extract_archive(
        output_dir=output_dir,
        archive_fileobj=task_plugin_archive,
    )

    expected_extracted_files: List[str] = [
        str(output_dir / x) for x in ["__init__.py", "plugin_module.py"]
    ]

    assert set(extracted_files) == set(expected_extracted_files)


def test_sanitize_file_path(io_file_service: IOFileService) -> None:
    clean_file_path1 = io_file_service.sanitize_file_path(
        filepath="dir/subdir/testfile.txt", path_prefix="/tmp"
    )
    clean_file_path2 = io_file_service.sanitize_file_path(
        filepath="dir/subdir/testfile.txt"
    )
    clean_file_path3 = io_file_service.sanitize_file_path(
        filepath="../testfile.txt", path_prefix="/tmp"
    )

    assert clean_file_path1 == Path("/tmp") / "testfile.txt"
    assert clean_file_path2 == Path.cwd() / "testfile.txt"
    assert clean_file_path3 == Path("/tmp") / "testfile.txt"


@pytest.mark.parametrize(
    "path, expected_result",
    [
        # assume fixed "/pre" prefix
        ("/foo/bar", "/pre/foo/bar"),
        ("foo/bar", "/pre/foo/bar"),
        ("/foo/../bar", None),
        ("foo/../bar", None),
    ],
)
def test_sanitize_file_path_preserve(
    io_file_service: IOFileService,
    path: str,
    expected_result: Optional[Union[str, Path]],
) -> None:
    actual_result = io_file_service.sanitize_file_path(path, "/pre", True)

    # wrap expected_result in a Path object, to make the below comparison work
    if expected_result is not None:
        expected_result = Path(expected_result)

    assert actual_result == expected_result


@pytest.mark.skipif(os.name != "nt", reason="requires Windows")
@pytest.mark.parametrize(
    "path, expected_result",
    [
        # assume fixed "/pre" prefix
        (r"C:\foo\bar", "/pre/foo/bar"),
        (r"C:foo\bar", "/pre/foo/bar"),
        (r"C:\foo\..\bar", None),
        (r"C:foo\..\bar", None),
    ],
)
def test_sanitize_file_path_preserve_windows(
    io_file_service: IOFileService,
    path: str,
    expected_result: Optional[Union[str, Path]],
) -> None:
    actual_result = io_file_service.sanitize_file_path(path, "/pre", True)

    # wrap expected_result in a Path object, to make the below comparison work
    if expected_result is not None:
        expected_result = Path(expected_result)

    assert actual_result == expected_result


def test_unsafe_tarball_path(tmp_path: Path, io_file_service: IOFileService) -> None:
    unsafe_tar_content = {"/a/b/c": b"123", "/a/../../b/c": b"456"}

    tarball_bytes = make_tarball_bytes(unsafe_tar_content)

    with pytest.raises(UnsafeArchiveMemberPath):
        io_file_service.safe_extract_archive(
            output_dir=tmp_path,
            archive_fileobj=io.BytesIO(tarball_bytes),
            preserve_paths=True,
        )



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import argparse
import logging
import logging.config
from typing import Union

import boto3

from dioptra.sdk.utilities.logging import (
    attach_stdout_stream_handler,
    configure_structlog,
)
from dioptra.worker.s3_download import s3_download


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="""
        Download file(s) from S3 to a directory.
        """
    )

    parser.add_argument(
        "-e",
        "--endpoint-url",
        help="""
        URL of an S3 endpoint to connect to
        """,
        required=True,
    )

    parser.add_argument(
        "-d",
        "--dest-dir",
        help="""
        The directory to download to.  Default: "%(default)s"
        """,
        default=".",
    )

    parser.add_argument(
        "-c",
        "--clear",
        help="""
        Clear the destination directory
        """,
        action="store_true",
    )

    parser.add_argument(
        "-p",
        "--preserve-paths",
        help="""
        Treat keys as paths and create that directory structure in the
        filesystem.  If not given, any key structure is flattened: directory
        structure is ignored, and filename path components (the last path
        component) are used to name the files created in the destination
        directory.
        """,
        action="store_true",
    )

    parser.add_argument(
        "-l",
        "--log-level",
        help="""
        Set the logging level.  "all" is a special setting which causes logging
        from all modules to display at the debug level, in addition to this
        application.  This can result in a lot of logging noise.
        Default: %(default)s
        """,
        choices=["all", "debug", "info", "warning", "error", "critical"],
        default="info",
    )

    parser.add_argument(
        "s3_uri",
        help="""
        S3 URI(s) to download from
        """,
        nargs="+",
    )

    return parser.parse_args()


def _setup_logging(log_level: Union[int, str] = logging.INFO) -> None:
    """
    Set up logging.

    Args:
        log_level: The logging level to use.  May be one of the integer log
            level constants or names recognized by the logging module, or
            "all".
    """
    if isinstance(log_level, str):
        log_level = log_level.upper()

    configure_structlog()

    if log_level == "ALL":
        log = logging.getLogger()
        log.setLevel(logging.DEBUG)
    else:
        log = logging.getLogger("dioptra")
        log.setLevel(log_level)

    attach_stdout_stream_handler(False, log)


def main() -> None:
    args = parse_args()
    _setup_logging(args.log_level)

    s3 = boto3.client("s3", endpoint_url=args.endpoint_url)

    s3_download(s3, args.dest_dir, args.clear, args.preserve_paths, *args.s3_uri)


if __name__ == "__main__":
    main()



"""empty message

Revision ID: 6a343ff9bc2f
Revises: 
Create Date: 2020-09-11 17:22:38.129031

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6a343ff9bc2f"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "experiments",
        sa.Column(
            "experiment_id",
            sa.BigInteger().with_variant(sa.Integer(), "sqlite"),
            nullable=False,
        ),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column("name", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("experiment_id"),
    )
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_experiments_name"), ["name"], unique=True)

    op.create_table(
        "jobs",
        sa.Column("job_id", sa.String(length=36), nullable=False),
        sa.Column("mlflow_run_id", sa.String(length=36), nullable=True),
        sa.Column("experiment_id", sa.BigInteger(), nullable=True),
        sa.Column("created_on", sa.DateTime(), nullable=True),
        sa.Column("last_modified", sa.DateTime(), nullable=True),
        sa.Column(
            "queue",
            sa.Enum("tensorflow_cpu", "tensorflow_gpu", name="jobqueue"),
            nullable=True,
        ),
        sa.Column("timeout", sa.Text(), nullable=True),
        sa.Column("workflow_uri", sa.Text(), nullable=True),
        sa.Column("entry_point", sa.Text(), nullable=True),
        sa.Column("entry_point_kwargs", sa.Text(), nullable=True),
        sa.Column("depends_on", sa.String(length=36), nullable=True),
        sa.Column(
            "status",
            sa.Enum(
                "queued", "started", "deferred", "finished", "failed", name="jobstatus"
            ),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["experiment_id"],
            ["experiments.experiment_id"],
        ),
        sa.PrimaryKeyConstraint("job_id"),
    )
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_jobs_experiment_id"), ["experiment_id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_jobs_mlflow_run_id"), ["mlflow_run_id"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_jobs_queue"), ["queue"], unique=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_jobs_queue"))
        batch_op.drop_index(batch_op.f("ix_jobs_mlflow_run_id"))
        batch_op.drop_index(batch_op.f("ix_jobs_experiment_id"))

    op.drop_table("jobs")
    with op.batch_alter_table("experiments", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_experiments_name"))

    op.drop_table("experiments")
    # ### end Alembic commands ###



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from pathlib import Path

import structlog
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def export_plugin_files(
    entry_point_plugin_files: list[models.EntryPointPluginFile],
    plugins_base_dir: Path,
    logger: BoundLogger | None = None,
) -> list[Path]:
    """Export an entrypoint's plugin files and folders in a specified directory.

    Args:
        entry_point_plugin_files: A list of entrypoint plugin files to be exported.
        plugins_base_dir: The base directory to save the plugin files and folders in.
        logger: A structlog logger object.

    Returns:
        A list of paths pointing to the exported plugin directories.
    """
    log = logger or LOGGER.new()  # noqa: F841
    plugin_dirs: set[Path] = set()

    for entry_point_plugin_file in entry_point_plugin_files:
        plugin = entry_point_plugin_file.plugin
        plugin_file = entry_point_plugin_file.plugin_file

        plugin_path = plugins_base_dir / plugin.name
        plugin_dirs.add(plugin_path)

        plugin_path.mkdir(parents=True, exist_ok=True)
        (plugin_path / plugin_file.filename).parent.mkdir(parents=True, exist_ok=True)
        (plugin_path / plugin_file.filename).write_text(plugin_file.contents or "")

    return list(plugin_dirs)



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The module defining the endpoints for Snapshots."""
import uuid
from typing import Any, Callable, Type, cast
from urllib.parse import unquote

import structlog
from flask import request
from flask_accepts import accepts, responds
from flask_login import login_required
from flask_restx import Namespace, Resource
from injector import ClassAssistedBuilder, inject
from marshmallow import Schema
from structlog.stdlib import BoundLogger

from dioptra.restapi.db import models
from dioptra.restapi.v1 import utils
from dioptra.restapi.v1.schemas import ResourceGetQueryParameters

from .service import ResourceSnapshotsIdService, ResourceSnapshotsService

LOGGER: BoundLogger = structlog.stdlib.get_logger()


def generate_resource_snapshots_endpoint(
    api: Namespace,
    resource_model: Type[models.ResourceSnapshot],
    resource_name: str,
    route_prefix: str,
    searchable_fields: dict[str, Any],
    page_schema: Type[Schema],
    build_fn: Callable,
) -> Resource:
    """Generates a ResourceSnapshotsEndpoint class.

    Args:
        api: The API
        resource_model: The ORM class for the resource snapshot.
        resource_name: The name of the resource.
        route_prefix: The prefix to append to the API URL.
        searchable_fields: A dictionary where the keys are the fields that can be
            searched and the values control how the query is constructed in the where
            clause.
        page_schema: The Marshmallow schema for the page response.
        build_fn: A function that builds the response object.

    Returns:
        The parameterized ResourceSnapshotsEndpoint class.
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(page_schema, Schema):
        model_name = "Snapshots" + "".join(
            page_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "Snapshots" + "".join(page_schema.__name__.rsplit("Schema", 1))

    @api.route("/<int:id>/snapshots")
    @api.param("id", f"ID for the {resource_name} resource.")
    class ResourceSnapshotsEndpoint(Resource):
        @inject
        def __init__(
            self,
            snapshots_service: ClassAssistedBuilder[ResourceSnapshotsService],
            *args,
            **kwargs,
        ) -> None:
            self._snapshots_service = snapshots_service.build(
                resource_model=resource_model,
                resource_type=resource_name,
                searchable_fields=searchable_fields,
            )
            super().__init__(*args, **kwargs)

        @login_required
        @accepts(query_params_schema=ResourceGetQueryParameters, api=api)
        @responds(schema=page_schema, model_name=model_name, api=api)
        def get(self, id: int):
            """Gets the Snapshots for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Snapshots", request_type="GET"
            )

            parsed_query_params = request.parsed_query_params  # type: ignore
            search_string = unquote(parsed_query_params["search"])
            page_index = parsed_query_params["index"]
            page_length = parsed_query_params["page_length"]

            snapshots, total_num_snapshots = cast(
                tuple[list[dict[str, Any]], int],
                self._snapshots_service.get(
                    resource_id=id,
                    search_string=search_string,
                    page_index=page_index,
                    page_length=page_length,
                    error_if_not_found=True,
                    log=log,
                ),
            )
            return utils.build_paging_envelope(
                f"{route_prefix}/{id}/snapshots",
                build_fn=build_fn,
                data=snapshots,
                group_id=None,
                query=search_string,
                draft_type=None,
                index=page_index,
                length=page_length,
                total_num_elements=total_num_snapshots,
            )

    return ResourceSnapshotsEndpoint


def generate_resource_snapshots_id_endpoint(
    api: Namespace,
    resource_model: Type[models.ResourceSnapshot],
    resource_name: str,
    response_schema: Type[Schema],
    build_fn: Callable,
) -> Resource:
    """Generates a ResourceSnapshotsIdEndpoint class.

    Args:
        api: The API
        resource_model: The ORM class for the resource snapshot.
        resource_name: The name of the resource.
        response_schema: The Marshmallow schema for the response.
        build_fn: A function that builds the response object.

    Returns:
        The parameterized ResourceSnapshotsIdEndpoint class.
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(response_schema, Schema):
        model_name = "SnapshotsId" + "".join(
            response_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "SnapshotsId" + "".join(
            response_schema.__name__.rsplit("Schema", 1)
        )

    @api.route("/<int:id>/snapshots/<int:snapshotId>")
    @api.param("id", f"ID for the {resource_name} resource.")
    @api.param("snapshotId", f"Snapshot ID for the {resource_name} resource.")
    class ResourcesSnapshotsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            snapshots_id_service: ClassAssistedBuilder[ResourceSnapshotsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._snapshots_id_service = snapshots_id_service.build(
                resource_model=resource_model,
                resource_type=resource_name,
            )
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=response_schema, model_name=model_name, api=api)
        def get(self, id: int, snapshotId: int):
            """Gets a Snapshot for the resource by snapshot id."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Snapshots", request_type="GET"
            )
            snapshot = cast(
                models.ResourceSnapshot,
                self._snapshots_id_service.get(
                    id, snapshot_id=snapshotId, error_if_not_found=True, log=log
                ),
            )
            return build_fn(snapshot)

    return ResourcesSnapshotsIdEndpoint


def generate_nested_resource_snapshots_endpoint(
    api: Namespace,
    resource_model: Type[models.ResourceSnapshot],
    resource_name: str,
    resource_route: str,
    base_resource_route: str,
    searchable_fields: dict[str, Any],
    page_schema: Type[Schema],
    build_fn: Callable,
) -> Resource:
    """Generates a ResourceSnapshotsEndpoint class.

    Args:
        api: The API
        resource_model: The ORM class for the resource snapshot.
        resource_name: The name of the resource.
        route_prefix: The prefix to append to the API URL.
        searchable_fields: A dictionary where the keys are the fields that can be
            searched and the values control how the query is constructed in the where
            clause.
        page_schema: The Marshmallow schema for the page response.
        build_fn: A function that builds the response object.

    Returns:
        The parameterized ResourceSnapshotsEndpoint class.
    """

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(page_schema, Schema):
        model_name = "NestedSnapshots" + "".join(
            page_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "NestedSnapshots" + "".join(
            page_schema.__name__.rsplit("Schema", 1)
        )

    route_singular = resource_route[:-1]
    resource_id = f"{route_singular}Id"

    @api.route(f"/<int:id>/{resource_route}/<int:{resource_id}>/snapshots")
    @api.param("id", f"ID for the {resource_name}.")
    @api.param(f"{resource_id}", f"ID for the {resource_name}.")
    class ResourceSnapshotsEndpoint(Resource):
        @inject
        def __init__(
            self,
            snapshots_service: ClassAssistedBuilder[ResourceSnapshotsService],
            *args,
            **kwargs,
        ) -> None:
            self._snapshots_service = snapshots_service.build(
                resource_model=resource_model,
                resource_type=resource_name,
                searchable_fields=searchable_fields,
            )
            super().__init__(*args, **kwargs)

        @login_required
        @accepts(query_params_schema=ResourceGetQueryParameters, api=api)
        @responds(schema=page_schema, model_name=model_name, api=api)
        def get(self, id: int, **kwargs):
            """Gets the Snapshots for the resource."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Snapshots", request_type="GET"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            parsed_query_params = request.parsed_query_params  # type: ignore
            search_string = unquote(parsed_query_params["search"])
            page_index = parsed_query_params["index"]
            page_length = parsed_query_params["page_length"]

            snapshots, total_num_snapshots = cast(
                tuple[list[models.ResourceSnapshot], int],
                self._snapshots_service.get(
                    resource_id=kwargs[resource_id],
                    search_string=search_string,
                    page_index=page_index,
                    page_length=page_length,
                    error_if_not_found=True,
                    log=log,
                ),
            )
            return utils.build_paging_envelope(
                f"{base_resource_route}/{id}/"
                f"{resource_route}/{kwargs[resource_id]}/snapshots",
                build_fn=build_fn,
                data=snapshots,
                group_id=None,
                query=search_string,
                draft_type=None,
                index=page_index,
                length=page_length,
                total_num_elements=total_num_snapshots,
            )

    return ResourceSnapshotsEndpoint


def generate_nested_resource_snapshots_id_endpoint(
    api: Namespace,
    resource_model: Type[models.ResourceSnapshot],
    resource_name: str,
    resource_route: str,
    response_schema: Type[Schema],
    build_fn: Callable,
) -> Resource:
    """Generates a ResourceSnapshotsIdEndpoint class.

    Args:
        api: The API
        resource_model: The ORM class for the resource snapshot.
        resource_name: The name of the resource.
        response_schema: The Marshmallow schema for the response.
        build_fn: A function that builds the response object.

    Returns:
        The parameterized ResourceSnapshotsIdEndpoint class.
    """

    route_singular = resource_route[:-1]
    resource_id = f"{route_singular}Id"

    # Based on: https://github.com/apryor6/flask_accepts/blob/05567461c421a534d6fc6e122d5e086b0b0e53aa/flask_accepts/utils.py#L154-L160  # noqa: B950
    if isinstance(response_schema, Schema):
        model_name = "NestedSnapshotsId" + "".join(
            response_schema.__class__.__name__.rsplit("Schema", 1)
        )
    else:
        model_name = "NestedSnapshotsId" + "".join(
            response_schema.__name__.rsplit("Schema", 1)
        )

    @api.route(
        f"/<int:id>/{resource_route}/<int:{resource_id}>/snapshots/<int:snapshotId>"
    )
    @api.param("id", f"ID for the {resource_name}.")
    @api.param(f"{resource_id}", f"ID for the {resource_name}.")
    @api.param("snapshotId", "ID for the Snapshot.")
    class ResourcesSnapshotsIdEndpoint(Resource):
        @inject
        def __init__(
            self,
            snapshots_id_service: ClassAssistedBuilder[ResourceSnapshotsIdService],
            *args,
            **kwargs,
        ) -> None:
            self._snapshots_id_service = snapshots_id_service.build(
                resource_model=resource_model,
                resource_type=resource_name,
            )
            super().__init__(*args, **kwargs)

        @login_required
        @responds(schema=response_schema, model_name=model_name, api=api)
        def get(self, id: int, snapshotId: int, **kwargs):
            """Gets a Snapshot for the resource by snapshot id."""
            log = LOGGER.new(
                request_id=str(uuid.uuid4()), resource="Snapshots", request_type="GET"
            )
            if set(kwargs.keys()) != set([resource_id]):
                unexpected_kwargs = {
                    k: v for k, v in kwargs.items() if resource_id != k
                }
                log.error("Unexpected input", kwargs=unexpected_kwargs)
                raise TypeError(
                    "delete method received unexpected keyword arguments: "
                    f"{list(unexpected_kwargs.keys())}"
                )

            snapshot = cast(
                models.ResourceSnapshot,
                self._snapshots_id_service.get(
                    kwargs[resource_id],
                    snapshot_id=snapshotId,
                    error_if_not_found=True,
                    log=log,
                ),
            )
            return build_fn(snapshot)

    return ResourcesSnapshotsIdEndpoint



"""Update entry point parameter types list

Revision ID: 6a75ede23821
Revises: 4b2d781f8bb4
Create Date: 2024-09-10 16:32:40.707231

"""

from typing import Annotated, Optional

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
    sessionmaker,
)

# revision identifiers, used by Alembic.
revision = "6a75ede23821"
down_revision = "4b2d781f8bb4"
branch_labels = None
depends_on = None


# Upgrade and downgrade inserts and deletes
UPGRADE_INSERTS = ["boolean", "integer", "list", "mapping"]
UPGRADE_DELETES = ["path", "uri"]
DOWNGRADE_INSERTS = ["path", "uri"]


# Migration data models
intpk = Annotated[
    int,
    mapped_column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True),
]
text_ = Annotated[str, mapped_column(sa.Text())]
bool_ = Annotated[bool, mapped_column(sa.Boolean())]
optionalstr = Annotated[Optional[str], mapped_column(sa.Text(), nullable=True)]


class UpgradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class DowngradeBase(DeclarativeBase, MappedAsDataclass):
    pass


class EntryPointParameterTypeUpgrade(UpgradeBase):
    __tablename__ = "entry_point_parameter_types"

    parameter_type: Mapped[text_] = mapped_column(primary_key=True)


class EntryPointParameterUpgrade(UpgradeBase):
    __tablename__ = "entry_point_parameters"

    entry_point_resource_snapshot_id: Mapped[intpk] = mapped_column(init=False)
    parameter_number: Mapped[intpk]
    parameter_type: Mapped[text_] = mapped_column(nullable=False)
    name: Mapped[text_] = mapped_column(nullable=False)
    default_value: Mapped[optionalstr]


class EntryPointParameterTypeDowngrade(DowngradeBase):
    __tablename__ = "entry_point_parameter_types"

    parameter_type: Mapped[text_] = mapped_column(primary_key=True)


def upgrade():
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with Session() as session:
        for parameter_type in UPGRADE_INSERTS:
            stmt = sa.select(EntryPointParameterTypeUpgrade).where(
                EntryPointParameterTypeUpgrade.parameter_type == parameter_type
            )

            if session.scalar(stmt) is None:
                session.add(
                    EntryPointParameterTypeUpgrade(parameter_type=parameter_type)
                )

        # Search for any parameters that are of type "path" or "uri" and convert them to
        # "string"
        to_string_params_stmt = sa.select(EntryPointParameterUpgrade).where(
            EntryPointParameterUpgrade.parameter_type.in_(["path", "uri"])
        )

        for entry_point_parameter in session.execute(to_string_params_stmt):
            entry_point_parameter.parameter_type = "string"

        for parameter_type in UPGRADE_DELETES:
            stmt = sa.select(EntryPointParameterTypeUpgrade).where(
                EntryPointParameterTypeUpgrade.parameter_type == parameter_type
            )
            entry_point_parameter_type = session.scalar(stmt)

            if entry_point_parameter_type is not None:
                session.delete(entry_point_parameter_type)

        session.commit()


def downgrade():
    bind = op.get_bind()
    Session = sessionmaker(bind=bind)

    with Session() as session:
        for parameter_type in DOWNGRADE_INSERTS:
            stmt = sa.select(EntryPointParameterTypeDowngrade).where(
                EntryPointParameterTypeDowngrade.parameter_type == parameter_type
            )

            if session.scalar(stmt) is None:
                session.add(
                    EntryPointParameterTypeDowngrade(parameter_type=parameter_type)
                )

        session.commit()



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
from __future__ import annotations

import os
from pathlib import Path
from posixpath import join as urljoin
from typing import Any, cast
from urllib.parse import urlparse, urlunparse

import requests


class DioptraClient(object):
    """Connects to the Dioptra REST api, and provides access to endpoints.

    Args:
        address: Address of the Dioptra REST api or if no address is given the
            DIOPTRA_RESTAPI_URI environment variable is used.
        api_version: The version of the Dioptra REST API to use. Defaults to "v0".

    Notes:
        See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html for
        more information on Dioptra's REST api.
    """

    def __init__(self, address: str | None = None, api_version: str = "v0") -> None:
        address = (
            f"{address}/api/{api_version}"
            if address
            else f"{os.environ['DIOPTRA_RESTAPI_URI']}/api/{api_version}"
        )
        self._scheme, self._netloc, self._path, _, _, _ = urlparse(address)

    @property
    def experiment_endpoint(self) -> str:
        """Experiment endpoint url"""
        return urlunparse(
            (self._scheme, self._netloc, urljoin(self._path, "experiment/"), "", "", "")
        )

    @property
    def job_endpoint(self) -> str:
        """Job endpoint url"""
        return urlunparse(
            (self._scheme, self._netloc, urljoin(self._path, "job/"), "", "", "")
        )

    @property
    def task_plugin_endpoint(self) -> str:
        """Task plugins endpoint url"""
        return urlunparse(
            (self._scheme, self._netloc, urljoin(self._path, "taskPlugin/"), "", "", "")
        )

    @property
    def task_plugin_builtins_endpoint(self) -> str:
        """Builtin task plugins endpoint url"""
        return urlunparse(
            (
                self._scheme,
                self._netloc,
                urljoin(self._path, "taskPlugin/dioptra_builtins"),
                "",
                "",
                "",
            )
        )

    @property
    def task_plugin_custom_endpoint(self) -> str:
        """Custom task plugins endpoint url"""
        return urlunparse(
            (
                self._scheme,
                self._netloc,
                urljoin(self._path, "taskPlugin/dioptra_custom"),
                "",
                "",
                "",
            )
        )

    @property
    def queue_endpoint(self) -> str:
        """Queue endpoint url"""
        return urlunparse(
            (self._scheme, self._netloc, urljoin(self._path, "queue/"), "", "", "")
        )

    def delete_custom_task_plugin(self, name: str) -> dict[str, Any]:
        """Deletes a custom task plugin by its unique name.

        Args:
            name: A unique string identifying a task plugin package within
                dioptra_custom collection.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'collection': 'dioptra_custom',
                    'status': 'Success',
                    'taskPluginName': ['evaluation']
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        plugin_name_query: str = urljoin(self.task_plugin_custom_endpoint, name)
        result = cast(dict[str, Any], requests.delete(plugin_name_query).json())
        return result

    def get_experiment_by_id(self, id: int) -> dict[str, Any]:
        """Gets an experiment by its unique identifier.

        Args:
            id: An integer identifying a registered experiment.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'lastModified': '2023-06-22T13:42:35.379462',
                    'experimentId': 10,
                    'name': 'mnist_feature_squeezing',
                    'createdOn': '2023-06-22T13:42:35.379462'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        experiment_id_query: str = urljoin(self.experiment_endpoint, str(id))
        return cast(dict[str, Any], requests.get(experiment_id_query).json())

    def get_experiment_by_name(self, name: str) -> dict[str, Any]:
        """Gets an experiment by its unique name.

        Args:
            name: The name of the experiment.

        Returns:
            The Dioptra REST api's response containing the experiment id, name, and
            metadata.

            Example::

                {
                    'experimentId': 1,
                    'name': 'mnist',
                    'createdOn': '2023-06-22T13:42:35.379462',
                    'lastModified': '2023-06-22T13:42:35.379462'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        experiment_name_query: str = urljoin(self.experiment_endpoint, "name", name)
        return cast(dict[str, Any], requests.get(experiment_name_query).json())

    def get_job_by_id(self, id: str) -> dict[str, Any]:
        """Gets a job by its unique identifier.

        Args:
            id: A string specifying a job's UUID.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'mlflowRunId': None,
                    'lastModified': '2023-06-26T15:26:43.100093',
                    'experimentId': 10,
                    'queueId': 2,
                    'workflowUri': 's3://workflow/268a7620/workflows.tar.gz',
                    'entryPoint': 'train',
                    'dependsOn': None,
                    'status': 'queued',
                    'timeout': '24h',
                    'jobId': '4eb2305e-57c3-4867-a59f-1a1ecd2033d4',
                    'entryPointKwargs': '-P model_architecture=shallow_net -P epochs=3',
                    'createdOn': '2023-06-26T15:26:43.100093'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        job_id_query: str = urljoin(self.job_endpoint, id)
        return cast(dict[str, Any], requests.get(job_id_query).json())

    def get_queue_by_id(self, id: int) -> dict[str, Any]:
        """Gets a queue by its unique identifier.

        Args:
            id: An integer identifying a registered queue.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'lastModified': '2023-04-24T20:53:09.801442',
                    'name': 'tensorflow_cpu',
                    'queueId': 1,
                    'createdOn': '2023-04-24T20:53:09.801442'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        queue_id_query: str = urljoin(self.queue_endpoint, str(id))
        return cast(dict[str, Any], requests.get(queue_id_query).json())

    def get_queue_by_name(self, name: str) -> dict[str, Any]:
        """Gets a queue by its unique name.

        Args:
            name: The name of the queue.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'lastModified': '2023-04-24T20:53:09.801442',
                    'name': 'tensorflow_cpu',
                    'queueId': 1,
                    'createdOn': '2023-04-24T20:53:09.801442'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        queue_name_query: str = urljoin(self.queue_endpoint, "name", name)
        return cast(dict[str, Any], requests.get(queue_name_query).json())

    def get_builtin_task_plugin(self, name: str) -> dict[str, Any]:
        """Gets a custom builtin plugin by its unique name.

        Args:
            name: A unique string identifying a task plugin package within
                dioptra_builtins collection.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'taskPluginName': 'attacks',
                    'collection': 'dioptra_builtins',
                    'modules': ['__init__.py', 'fgm.py']
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        task_plugin_name_query: str = urljoin(self.task_plugin_builtins_endpoint, name)
        return cast(dict[str, Any], requests.get(task_plugin_name_query).json())

    def get_custom_task_plugin(self, name: str) -> dict[str, Any]:
        """Gets a custom task plugin by its unique name.

        Args:
            name: A unique string identifying a task plugin package within
                dioptra_builtins collection.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'taskPluginName': 'custom_poisoning_plugins',
                    'collection': 'dioptra_custom',
                    'modules': [
                        '__init__.py',
                        'attacks_poison.py',
                        'data_tensorflow.py',
                        'datasetup.py',
                        'defenses_image_preprocessing.py',
                        'defenses_training.py',
                        'estimators_keras_classifiers.py',
                        'registry_art.py',
                        'tensorflow.py'
                    ]
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        task_plugin_name_query: str = urljoin(self.task_plugin_custom_endpoint, name)
        return cast(dict[str, Any], requests.get(task_plugin_name_query).json())

    def list_experiments(self) -> list[dict[str, Any]]:
        """Gets a list of all registered experiments.

        Returns:
            A list of responses detailing all experiments.

            Example::

                [
                    {
                        'lastModified': '2023-04-24T20:20:27.315687',
                        'experimentId': 1,
                        'name': 'mnist',
                        'createdOn': '2023-04-24T20:20:27.315687'
                    },
                    ...
                    {
                        'lastModified': '2023-06-22T13:42:35.379462',
                        'experimentId': 10,
                        'name': 'mnist_feature_squeezing',
                        'createdOn': '2023-06-22T13:42:35.379462'
                    }
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        return cast(list[dict[str, Any]], requests.get(self.experiment_endpoint).json())

    def list_jobs(self) -> list[dict[str, Any]]:
        """Gets a list of all submitted jobs.

        Returns:
            A list of responses detailing all jobs.

            Example::

                [
                    {
                        'mlflowRunId': None,
                        'lastModified': '2023-04-24T20:54:30.722304',
                        'experimentId': 2,
                        'queueId': 2,
                        'workflowUri': 's3://workflow/268a7620/workflows.tar.gz',
                        'entryPoint': 'train',
                        'dependsOn': None,
                        'status': 'queued',
                        'timeout': '1h',
                        'jobId': 'a4c574dd-cbd1-43c9-9afe-17d69cd1c73d',
                        'entryPointKwargs': '-P data_dir=/nfs/data/Mnist',
                        'createdOn': '2023-04-24T20:54:30.722304'
                    },
                    ...
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        return cast(list[dict[str, Any]], requests.get(self.job_endpoint).json())

    def list_queues(self) -> list[dict[str, Any]]:
        """Gets a list of all registered queues.

        Returns:
            A list of responses detailing all registered queues.

            Example::

                [
                    {
                        'lastModified': '2023-04-24T20:53:09.801442',
                        'name': 'tensorflow_cpu',
                        'queueId': 1,
                        'createdOn': '2023-04-24T20:53:09.801442'
                    },
                    {
                        'lastModified': '2023-04-24T20:53:09.824101',
                        'name': 'tensorflow_gpu',
                        'queueId': 2,
                        'createdOn': '2023-04-24T20:53:09.824101'
                    },
                    {
                        'lastModified': '2023-04-24T20:53:09.867917',
                        'name': 'pytorch_cpu',
                        'queueId': 3,
                        'createdOn': '2023-04-24T20:53:09.867917'
                    },
                    {
                        'lastModified': '2023-04-24T20:53:09.893451',
                        'name': 'pytorch_gpu',
                        'queueId': 4,
                        'createdOn': '2023-04-24T20:53:09.893451'
                    }
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        return cast(list[dict[str, Any]], requests.get(self.queue_endpoint).json())

    def list_all_task_plugins(self) -> list[dict[str, Any]]:
        """Gets a list of all registered builtin task plugins.

        Returns:
            A list of responses detailing all plugins.

            Example::

                [
                    {
                        'taskPluginName': 'artifacts',
                        'collection': 'dioptra_builtins',
                        'modules': ['__init__.py', 'mlflow.py', 'utils.py']
                    },
                    ...
                    {
                        'taskPluginName': 'pixel_threshold',
                        'collection': 'dioptra_custom',
                        'modules': ['__init__.py', 'pixelthreshold.py']
                    }
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """

        return cast(
            list[dict[str, Any]], requests.get(self.task_plugin_endpoint).json()
        )

    def list_builtin_task_plugins(self) -> list[dict[str, Any]]:
        """Gets a list of all registered builtin task plugins.

        Returns:
            A list of responses detailing all builtin plugins.

            Example::

                [
                    {
                        'taskPluginName': 'artifacts',
                        'collection': 'dioptra_builtins',
                        'modules': ['__init__.py', 'mlflow.py', 'utils.py']
                    },
                    ...
                    {
                        'taskPluginName': 'backend_configs',
                        'collection': 'dioptra_builtins',
                        'modules': ['__init__.py', 'tensorflow.py']
                    }
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html for
            more information on Dioptra's REST api.
        """
        return cast(
            list[dict[str, Any]],
            requests.get(self.task_plugin_builtins_endpoint).json(),
        )

    def list_custom_task_plugins(self) -> list[dict[str, Any]]:
        """Gets a list of all registered custom task plugins.

        Returns:
            A list of responses detailing all custom plugins.

            Example::

                [
                    {
                        'taskPluginName': 'model_inversion',
                        'collection': 'dioptra_custom',
                        'modules': ['__init__.py', 'modelinversion.py']
                    },
                    ...
                    {
                        'taskPluginName': 'pixel_threshold',
                        'collection': 'dioptra_custom',
                        'modules': ['__init__.py', 'pixelthreshold.py']
                    }
                ]

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        return cast(
            list[dict[str, Any]], requests.get(self.task_plugin_custom_endpoint).json()
        )

    def lock_queue(self, name: str) -> dict[str, Any]:
        """Locks the queue (name reference) if it is unlocked.

        Args:
            name: The name of the queue.

        Returns:
            The Dioptra REST api's response.

            Example::

                {'name': ['tensorflow_cpu'], 'status': 'Success'}

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        queue_name_query: str = urljoin(self.queue_endpoint, "name", name, "lock")
        return cast(dict[str, Any], requests.put(queue_name_query).json())

    def unlock_queue(self, name: str) -> dict[str, Any]:
        """Removes the lock from the queue (name reference) if it exists.

        Args:
            name: The name of the queue.

        Returns:
            The Dioptra REST api's response.

            Example::

                {'name': ['tensorflow_cpu'], 'status': 'Success'}

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        queue_name_query: str = urljoin(self.queue_endpoint, "name", name, "lock")
        return cast(dict[str, Any], requests.delete(queue_name_query).json())

    def register_experiment(self, name: str) -> dict[str, Any]:
        """Creates a new experiment via an experiment registration form.

        Args:
            name: The name to register as a new experiment.

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'lastModified': '2023-06-26T15:45:09.232878',
                    'experimentId': 11,
                    'name': 'experiment1234',
                    'createdOn': '2023-06-26T15:45:09.232878'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        experiment_registration_form = {"name": name}

        response = requests.post(
            self.experiment_endpoint,
            json=experiment_registration_form,
        )

        return cast(dict[str, Any], response.json())

    def register_queue(self, name: str = "tensorflow_cpu") -> dict[str, Any]:
        """Creates a new queue via a queue registration form.

        Args:
            name: The name to register as a new queue. Defaults to "tensorflow_cpu".

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'lastModified': '2023-06-26T15:48:47.662293',
                    'name': 'queue',
                    'queueId': 7,
                    'createdOn': '2023-06-26T15:48:47.662293'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        queue_registration_form = {"name": name}

        response = requests.post(
            self.queue_endpoint,
            json=queue_registration_form,
        )

        return cast(dict[str, Any], response.json())

    def submit_job(
        self,
        workflows_file: str | Path,
        experiment_name: str,
        entry_point: str,
        entry_point_kwargs: str | None = None,
        depends_on: str | None = None,
        queue: str = "tensorflow_cpu",
        timeout: str = "24h",
    ) -> dict[str, Any]:
        """Creates a new job via a job submission form with an attached file.

        Args:
            workflows_file: A tarball archive or zip file containing, at a minimum,
                a MLproject file and its associated entry point scripts.
            experiment_name:The name of a registered experiment.
            entry_point: Entrypoint name.
            entry_point_kwargs: A string listing parameter values to pass to the
                entry point for the job. The list of parameters is specified using the
                following format: “-P param1=value1 -P param2=value2”. Defaults to None.
            depends_on: A UUID for a previously submitted job to set as a dependency
                for the current job. Defaults to None.
            queue: Name of the queue the job is submitted to. Defaults to
                "tensorflow_cpu".
            timeout: The maximum alloted time for a job before it times out and is
                stopped. Defaults to "24h".

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'createdOn': '2023-06-26T15:26:43.100093',
                    'dependsOn': None,
                    'entryPoint': 'train',
                    'entryPointKwargs': '-P data_dir=/dioptra/data/Mnist',
                    'experimentId': 10,
                    'jobId': '4eb2305e-57c3-4867-a59f-1a1ecd2033d4',
                    'lastModified': '2023-06-26T15:26:43.100093',
                    'mlflowRunId': None,
                    'queueId': 2,
                    'status': 'queued',
                    'timeout': '24h',
                    'workflowUri': 's3://workflow/07d2c0a9/workflows.tar.gz'
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        job_form: dict[str, Any] = {
            "experimentName": experiment_name,
            "queue": queue,
            "timeout": timeout,
            "entryPoint": entry_point,
        }

        if entry_point_kwargs is not None:
            job_form["entryPointKwargs"] = entry_point_kwargs

        if depends_on is not None:
            job_form["dependsOn"] = depends_on

        workflows_file = Path(workflows_file)

        with workflows_file.open("rb") as f:
            job_files = {"workflow": (workflows_file.name, f)}
            response = requests.post(
                self.job_endpoint,
                data=job_form,
                files=job_files,
            )

        return cast(dict[str, Any], response.json())

    def upload_custom_plugin_package(
        self,
        custom_plugin_name: str,
        custom_plugin_file: str | Path,
        collection: str = "dioptra_custom",
    ) -> dict[str, Any]:
        """Registers a new task plugin uploaded via the task plugin upload form.

        Args:
            custom_plugin_name: Plugin name for the upload form.
            custom_plugin_file: Path to custom plugin.
            collection: Collection to upload the plugin to. Defaults to
                "dioptra_custom".

        Returns:
            The Dioptra REST api's response.

            Example::

                {
                    'taskPluginName': 'evaluation',
                    'collection': 'dioptra_custom',
                    'modules': [
                        'tensorflow.py',
                        'import_keras.py',
                        '__init__.py'
                    ]
                }

        Notes:
            See https://pages.nist.gov/dioptra/user-guide/api-reference-restapi.html
            for more information on Dioptra's REST api.
        """
        plugin_upload_form = {
            "taskPluginName": custom_plugin_name,
            "collection": collection,
        }

        custom_plugin_file = Path(custom_plugin_file)

        with custom_plugin_file.open("rb") as f:
            custom_plugin_file_dict = {"taskPluginFile": (custom_plugin_file.name, f)}
            response = requests.post(
                self.task_plugin_endpoint,
                data=plugin_upload_form,
                files=custom_plugin_file_dict,
            )

        return cast(dict[str, Any], response.json())



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import datetime
from typing import TYPE_CHECKING

from sqlalchemy import Column, ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from dioptra.restapi.db.db import datetimetz, db, intpk, text_

if TYPE_CHECKING:
    from .groups import Group
    from .resources import Resource
    from .users import User

# -- Tables (no ORM) -------------------------------------------------------------------

user_lock_types_table = db.Table(
    "user_lock_types",
    Column("user_lock_type", Text(), primary_key=True),
)
group_lock_types_table = db.Table(
    "group_lock_types",
    Column("group_lock_type", Text(), primary_key=True),
)
resource_lock_types_table = db.Table(
    "resource_lock_types",
    Column("resource_lock_type", Text(), primary_key=True),
)

# -- ORM Classes -----------------------------------------------------------------------


class UserLock(db.Model):  # type: ignore[name-defined]
    __tablename__ = "user_locks"

    # Database fields
    user_id: Mapped[intpk] = mapped_column(ForeignKey("users.user_id"), init=False)
    user_lock_type: Mapped[text_] = mapped_column(
        ForeignKey("user_lock_types.user_lock_type"), primary_key=True
    )
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    user: Mapped["User"] = relationship("User", back_populates="locks")

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class GroupLock(db.Model):  # type: ignore[name-defined]
    __tablename__ = "group_locks"

    # Database fields
    group_id: Mapped[intpk] = mapped_column(ForeignKey("groups.group_id"), init=False)
    group_lock_type: Mapped[text_] = mapped_column(
        ForeignKey("group_lock_types.group_lock_type"), primary_key=True
    )
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    group: Mapped["Group"] = relationship("Group", back_populates="locks")

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp


class ResourceLock(db.Model):  # type: ignore[name-defined]
    __tablename__ = "resource_locks"

    # Database fields
    resource_id: Mapped[intpk] = mapped_column(
        ForeignKey("resources.resource_id"), init=False
    )
    resource_lock_type: Mapped[text_] = mapped_column(
        ForeignKey("resource_lock_types.resource_lock_type"), primary_key=True
    )
    created_on: Mapped[datetimetz] = mapped_column(init=False, nullable=False)

    # Relationships
    resource: Mapped["Resource"] = relationship("Resource", back_populates="locks")

    # Initialize default values using dataclass __post_init__ method
    # https://docs.python.org/3/library/dataclasses.html#dataclasses.__post_init__
    def __post_init__(self) -> None:
        timestamp = datetime.datetime.now(tz=datetime.timezone.utc)
        self.created_on = timestamp



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
"""The schemas for serializing/deserializing Model resources."""
from marshmallow import Schema, fields

from dioptra.restapi.v1.artifacts.schema import ArtifactRefSchema
from dioptra.restapi.v1.schemas import (
    BasePageSchema,
    GroupIdQueryParametersSchema,
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
    generate_base_resource_ref_schema,
    generate_base_resource_schema,
)

ModelBaseRefSchema = generate_base_resource_ref_schema("Model")


class ModelRefSchema(ModelBaseRefSchema):  # type: ignore
    """The reference schema for the data stored in a Model resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Model resource."),
        required=True,
    )


class ModelVersionRefSchema(Schema):
    """The reference schema for the data stored in a Model Version."""

    versionNumber = fields.Integer(
        attribute="version_number",
        metadata=dict(description="The version number of the Model."),
        dump_only=True,
    )
    url = fields.Url(
        attribute="url",
        metadata=dict(description="URL for accessing the full Model Version."),
        relative=True,
    )


class ModelVersionMutableFieldsSchema(Schema):
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Model Version."),
        load_default=None,
    )


class ModelVersionSchema(ModelVersionMutableFieldsSchema):
    """The schema for the data stored in a ModelVersion resource."""

    model = fields.Nested(
        ModelRefSchema,
        attribute="model",
        metadata=dict(description="The Model resource."),
        dump_only=True,
    )
    artifactId = fields.Integer(
        attribute="artifact_id",
        data_key="artifact",
        metadata=dict(description="The artifact representing the Model Version."),
        load_only=True,
        required=True,
    )
    artifact = fields.Nested(
        ArtifactRefSchema,
        attribute="artifact",
        metadata=dict(description="The artifact representing the Model Version."),
        dump_only=True,
    )
    versionNumber = fields.Integer(
        attribute="version_number",
        metadata=dict(description="The version number of the Model."),
        dump_only=True,
    )
    createdOn = fields.DateTime(
        attribute="created_on",
        metadata=dict(description="Timestamp when the Model Version was created."),
        dump_only=True,
    )


class ModelMutableFieldsSchema(Schema):
    """The fields schema for the mutable data in a Model resource."""

    name = fields.String(
        attribute="name",
        metadata=dict(description="Name of the Model resource."),
        required=True,
    )
    description = fields.String(
        attribute="description",
        metadata=dict(description="Description of the Model resource."),
        load_default=None,
    )


ModelBaseSchema = generate_base_resource_schema("Model", snapshot=True)


class ModelSchema(ModelMutableFieldsSchema, ModelBaseSchema):  # type: ignore
    """The schema for the data stored in a Model resource."""

    versions = fields.Nested(
        ModelVersionRefSchema,
        many=True,
        attribute="versions",
        metadata=dict(description="The details of this model version."),
        dump_only=True,
    )
    latestVersion = fields.Nested(
        ModelVersionSchema,
        attribute="latest_version",
        metadata=dict(description="The details of latest version of this model."),
        dump_only=True,
    )


class ModelPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Model resource."""

    data = fields.Nested(
        ModelSchema,
        many=True,
        metadata=dict(description="List of Model resources in the current page."),
    )


class ModelVersionPageSchema(BasePageSchema):
    """The paged schema for the data stored in a Model resource."""

    data = fields.Nested(
        ModelVersionSchema,
        many=True,
        metadata=dict(description="List of Model resources in the current page."),
    )


class ModelGetQueryParameters(
    PagingQueryParametersSchema,
    GroupIdQueryParametersSchema,
    SearchQueryParametersSchema,
    SortByGetQueryParametersSchema,
):
    """The query parameters for the GET method of the /models endpoint."""


class ModelVersionGetQueryParameters(
    PagingQueryParametersSchema,
    SearchQueryParametersSchema,
):
    """The query parameters for the GET method of the /models/{id}/versions endpoint."""



# This Software (Dioptra) is being made available as a public service by the
# National Institute of Standards and Technology (NIST), an Agency of the United
# States Department of Commerce. This software was developed in part by employees of
# NIST and in part by NIST contractors. Copyright in portions of this software that
# were developed by NIST contractors has been licensed or assigned to NIST. Pursuant
# to Title 17 United States Code Section 105, works of NIST employees are not
# subject to copyright protection in the United States. However, NIST may hold
# international copyright in software created by its employees and domestic
# copyright (or licensing rights) in portions of software that were assigned or
# licensed to NIST. To the extent that NIST holds copyright in this software, it is
# being made available under the Creative Commons Attribution 4.0 International
# license (CC BY 4.0). The disclaimers of the CC BY 4.0 license apply to all parts
# of the software developed or licensed by NIST.
#
# ACCESS THE FULL CC BY 4.0 LICENSE HERE:
# https://creativecommons.org/licenses/by/4.0/legalcode
import textwrap
import uuid
from dataclasses import dataclass

from faker import Faker
from passlib.hash import pbkdf2_sha256

from dioptra.restapi.db import models


@dataclass
class FakeAccount(object):
    """Dataclass for storing a fake account's ORM objects.

    Attributes:
        user: An ORM object representing a fake user.
        group: An ORM object representing a fake group.
    """

    user: models.User
    group: models.Group


@dataclass
class FakePlugin(object):
    """Dataclass for storing a fake plugin's ORM objects.

    Attributes:
        plugin: An ORM object representing a fake plugin.
        plugin_task: An ORM object representing a fake plugin task.
        plugin_file: An ORM object representing a fake plugin file.
        init_plugin_file: An ORM object representing a fake init plugin file.
    """

    plugin: models.Plugin
    plugin_task: models.PluginTask
    plugin_file: models.PluginFile
    init_plugin_file: models.PluginFile


class FakeData(object):
    """A fake data generator for creating ORM objects for use in testing."""

    def __init__(self, faker: Faker) -> None:
        """Instantiate a new FakeData object.

        Args:
            faker: The Faker object to use for generating fake data.
        """
        self._faker = faker

    def account(self) -> FakeAccount:
        """Generate a fake account consisting of a user that belongs to a default group.

        Returns:
            A FakeAccount object containing user and group ORM objects.
        """
        username = self._faker.user_name()
        password = self._faker.password()
        hashed_password = pbkdf2_sha256.hash(password)
        email_address = self._faker.email()

        user = models.User(
            username=username, password=hashed_password, email_address=email_address
        )
        group = models.Group(name=username, creator=user)

        group.members.append(
            models.GroupMember(
                user=user,
                read=True,
                write=True,
                share_read=True,
                share_write=True,
            )
        )
        group.managers.append(models.GroupManager(user=user, owner=True, admin=True))

        return FakeAccount(user=user, group=group)

    def tag(self, creator: models.User, group: models.Group) -> models.Tag:
        """Generate a fake tag.

        Args:
            creator: The user that created the tag.
            group: The group that owns the tag.

        Returns:
            An ORM object representing a fake tag.
        """
        tag = models.Tag(name=self._faker.slug(), owner=group, creator=creator)
        return tag

    def queue(
        self,
        creator: models.User,
        group: models.Group,
        include_description: bool = False,
    ) -> models.Queue:
        """Generate a fake queue.

        Args:
            creator: The user that created the queue.
            group: The group that owns the queue.
            include_description: Whether to generate a fake description for the queue.

        Returns:
            An ORM object representing a fake queue.
        """
        queue_name = self._faker.sentence(nb_words=3, variable_nb_words=True)
        description = self._faker.sentence() if include_description else None

        resource = models.Resource(resource_type="queue", owner=group)
        queue = models.Queue(
            name=queue_name, description=description, resource=resource, creator=creator
        )

        return queue

    def experiment(
        self,
        creator: models.User,
        group: models.Group,
        include_description: bool = False,
    ) -> models.Experiment:
        """Generate a fake experiment.

        Args:
            creator: The user that created the experiment.
            group: The group that owns the experiment.
            include_description: Whether to generate a fake description for the
                experiment.

        Returns:
            An ORM object representing a fake experiment.
        """
        experiment_name = self._faker.word()
        description = self._faker.sentence() if include_description else None

        resource = models.Resource(resource_type="experiment", owner=group)
        experiment = models.Experiment(
            name=experiment_name,
            description=description,
            resource=resource,
            creator=creator,
        )

        return experiment

    def plugin_task_parameter_type(
        self,
        creator: models.User,
        group: models.Group,
        name: str | None = None,
        include_description: bool = False,
    ) -> models.Experiment:
        """Generate a fake plugin task parameter type.

        Args:
            creator: The user that created the plugin task parameter type.
            group: The group that owns the plugin task parameter type.
            name: The name of the plugin task parameter type. If not provided, a random
                name will be generated.
            include_description: Whether to generate a fake description for the plugin
                task parameter type.

        Returns:
            An ORM object representing a fake plugin task parameter type.
        """
        type_name = name or self._faker.word()
        structure = None
        description = self._faker.sentence() if include_description else None

        resource = models.Resource(
            resource_type="plugin_task_parameter_type", owner=group
        )
        plugin_task_parameter_type = models.PluginTaskParameterType(
            name=type_name,
            structure=structure,
            description=description,
            resource=resource,
            creator=creator,
        )

        return plugin_task_parameter_type

    def plugin(
        self,
        creator: models.User,
        group: models.Group,
        str_parameter_type: models.PluginTaskParameterType,
        include_description: bool = False,
    ) -> FakePlugin:
        """Generate a fake plugin.

        Args:
            creator: The user that created the plugin.
            group: The group that owns the plugin.
            str_parameter_type: The plugin task parameter type for string parameters.
            include_description: Whether to generate a fake description for the plugin.

        Returns:
            A FakePlugin object containing plugin, plugin task, plugin file, and init
            plugin file ORM objects.
        """
        plugin_name = self._faker.word()
        plugin_description = self._faker.sentence() if include_description else None
        plugin_init_file_name = "__init__.py"
        plugin_init_file_description = (
            self._faker.sentence() if include_description else None
        )
        plugin_file_stem = self._faker.word()
        plugin_file_name = f"{plugin_file_stem}.py"
        plugin_file_description = (
            self._faker.sentence() if include_description else None
        )
        plugin_file_contents = textwrap.dedent(
            """from dioptra import pyplugs

            @pyplugs.register
            def hello_world(name: str) -> str:
                return f"Hello, {name}!"
            """
        )
        input_parameter_name = "name"
        output_parameter_name = "hello_world_message"
        plugin_task_name = "hello_world"

        plugin_resource = models.Resource(resource_type="plugin", owner=group)
        init_plugin_file_resource = models.Resource(
            resource_type="plugin_file", owner=group
        )
        plugin_file_resource = models.Resource(resource_type="plugin_file", owner=group)

        plugin = models.Plugin(
            name=plugin_name,
            resource=plugin_resource,
            description=plugin_description,
            creator=creator,
        )
        init_plugin_file = models.PluginFile(
            filename=plugin_init_file_name,
            contents="",
            description=plugin_init_file_description,
            resource=init_plugin_file_resource,
            creator=creator,
        )
        plugin_file = models.PluginFile(
            filename=plugin_file_name,
            contents=plugin_file_contents,
            description=plugin_file_description,
            resource=plugin_file_resource,
            creator=creator,
        )
        input_parameter = models.PluginTaskInputParameter(
            name=input_parameter_name,
            parameter_number=1,
            parameter_type=str_parameter_type,
            required=True,
        )
        output_parameter = models.PluginTaskOutputParameter(
            name=output_parameter_name,
            parameter_number=1,
            parameter_type=str_parameter_type,
        )
        plugin_task = models.PluginTask(
            file=plugin_file,
            plugin_task_name=plugin_task_name,
            input_parameters=[input_parameter],
            output_parameters=[output_parameter],
        )

        init_plugin_file.parents.append(plugin.resource)
        plugin_file.parents.append(plugin.resource)

        return FakePlugin(
            plugin=plugin,
            plugin_task=plugin_task,
            plugin_file=plugin_file,
            init_plugin_file=init_plugin_file,
        )

    def entry_point(
        self,
        creator: models.User,
        group: models.Group,
        plugin: models.Plugin,
        plugin_files: list[models.PluginFile],
        queue: models.Queue,
        include_description: bool = False,
    ) -> models.EntryPoint:
        """Generate a fake entry point.

        Args:
            creator: The user that created the entry point.
            group: The group that owns the entry point.
            plugin: A plugin associated with the entry point.
            plugin_files: Plugin files associated with the entry point.
            queue: A queue associated with the entry point.
            include_description: Whether to generate a fake description for the entry
                point.

        Returns:
            An ORM object representing a fake entry point.
        """
        entry_point_name = self._faker.word()
        entry_point_parameters = [
            models.EntryPointParameter(
                parameter_number=1,
                parameter_type="string",
                name="name",
                default_value=self._faker.word(),
            ),
        ]
        task_graph = textwrap.dedent(
            """# hello world graph
            graph:
              message:
                hello_world: $name
            """
        )
        description = self._faker.sentence() if include_description else None
        entry_point_resource = models.Resource(resource_type="entry_point", owner=group)
        entry_point = models.EntryPoint(
            name=entry_point_name,
            task_graph=task_graph,
            parameters=entry_point_parameters,
            description=description,
            resource=entry_point_resource,
            creator=creator,
        )
        entry_point_plugin_files = [
            models.EntryPointPluginFile(
                entry_point=entry_point, plugin=plugin, plugin_file=plugin_file
            )
            for plugin_file in plugin_files
        ]
        entry_point.entry_point_plugin_files.extend(entry_point_plugin_files)
        entry_point.children.append(queue.resource)
        return entry_point

    def job(
        self,
        creator: models.User,
        group: models.Group,
        entry_point: models.EntryPoint,
        experiment: models.Experiment,
        queue: models.Queue,
        include_description: bool = False,
    ) -> models.Job:
        """Generate a fake job.

        Args:
            creator: The user that created the job.
            group: The group that owns the job.
            entry_point: The entry point used to instantiate the job.
            experiment: The experiment the job is for.
            queue: The queue where the job will be run.associated with the job.
            include_description: Whether to generate a fake description for the job.

        Returns:
            An ORM object representing a fake job.
        """
        timeout = "24h"
        status = "queued"
        description = self._faker.sentence() if include_description else None

        job_resource = models.Resource(resource_type="job", owner=group)
        entry_point_parameter_values = [
            models.EntryPointParameterValue(
                value=self._faker.word(),
                job_resource=job_resource,
                parameter=entry_point.parameters[0],
            ),
        ]
        job = models.Job(
            timeout=timeout,
            status=status,
            description=description,
            resource=job_resource,
            creator=creator,
        )
        job.entry_point_job = models.EntryPointJob(
            job_resource=job_resource,
            entry_point=entry_point,
            entry_point_parameter_values=entry_point_parameter_values,
        )
        job.experiment_job = models.ExperimentJob(
            job_resource=job_resource,
            experiment=experiment,
        )
        job.queue_job = models.QueueJob(
            job_resource=job_resource,
            queue=queue,
        )

        return job

    def job_mlflow_run(self, job: models.Job) -> models.JobMlflowRun:
        """Generate a fake job MLflow run.

        Args:
            job: The job that registered a run on the MLflow Tracking server.

        Returns:
            An ORM object representing a fake job MLflow run.
        """
        job_mlflow_run = models.JobMlflowRun(
            job_resource_id=job.resource_id,
            mlflow_run_id=uuid.uuid4(),
        )

        return job_mlflow_run

    def artifact(
        self,
        creator: models.User,
        group: models.Group,
        job: models.Job,
        include_description: bool = False,
    ) -> models.Artifact:
        """Generate a fake artifact.

        Args:
            creator: The user that created the artifact.
            group: The group that owns the artifact.
            job: The job that produced the artifact.
            include_description: Whether to generate a fake description for the
                artifact.

        Returns:
            An ORM object representing a fake artifact.
        """
        description = self._faker.sentence() if include_description else None
        artifact_resource = models.Resource(resource_type="artifact", owner=group)
        new_artifact = models.Artifact(
            uri=self._faker.uri(schemes=["s3"]),
            description=description,
            resource=artifact_resource,
            creator=creator,
        )

        # This is how to link an artifact to a job resource
        job.children.append(new_artifact.resource)

        return new_artifact

    def ml_model(
        self,
        creator: models.User,
        group: models.Group,
        artifact: models.Artifact,
        include_description: bool = False,
    ) -> models.MlModel:
        """Generate a fake ML model.

        Args:
            creator: The user that created the ML model.
            group: The group that owns the ML model.
            artifact: The artifact to register as a ML model.
            include_description: Whether to generate a fake description for the ML
                model.

        Returns:
            An ORM object representing a fake ML model.
        """
        description = self._faker.sentence() if include_description else None
        ml_model_resource = models.Resource(resource_type="ml_model", owner=group)
        new_ml_model = models.MlModel(
            name=self._faker.word(),
            description=description,
            resource=ml_model_resource,
            creator=creator,
            artifact=artifact,
        )

        return new_ml_model
